from collections.abc import Generator
from typing import Callable, Optional

from _typeshed import Incomplete

version: str

# def primegen(limit=...) -> Generator[Incomplete, Incomplete, Incomplete]: ...
# def ilog(x, b): ...

def modinv(a: int, m: int) -> int: ...
def introot(n: int, r: int = 2) -> int: ...
def ispower(n: int, r: int = 0) -> Optional[tuple[int, int] | int]: ...

# def jacobi(a, n): ...

def isprime(
    n: int,
    tb: tuple[int, ...] = ...,
) -> bool: ...

# def pollardrho_brent(n, verbose: bool = False): ...
# def pollard_pm1(n, B1: int = 100, B2: int = 1000, verbose: bool = False): ...
# def mlucas(v, a, n): ...
# def williams_pp1(n, verbose: bool = False): ...
# def ecadd(p1, p2, p0, n): ...
# def ecdub(p, A, n): ...
# def ecmul(m, p, A, n): ...
# def secm(n, B1, B2, seed): ...
# def ecmparams(n) -> Generator[Incomplete, None, None]: ...
# def ecm(n, paramseq=..., nprocs: int = 1, verbose: bool = False): ...

def sqrtmod_prime(a: int, p: int) -> int: ...

# def siqs(n, verbose: bool = False): ...
# def multifactor(n, methods=..., verbose: bool = False): ...

def primefac(
    n: int,
    trial: int = 1000,
    rho: int = 42000,
    verbose: bool = False,
    methods: Callable[[int, Incomplete], int] = ...,
) -> Generator[int, None, None]: ...

# def rpn(instr) -> Generator[Incomplete, None, None]: ...

usage: str



import sys

import pytest
from toy_crypto import bit_utils


@pytest.fixture
def sets() -> dict[str, set[int]]:
    sets: dict[str, set[int]] = dict()
    sets["universe"] = set(range(100))

    sets["primes"] = {
        2, 3, 5, 7, 11, 13, 17, 19,
        23, 29, 31, 37, 41, 43, 47,
        53, 59, 61, 67, 71, 73,
        79, 83, 89, 97,
    }  # fmt: skip

    sets.update(
        {f"multiples_{n}": set(range(0, 100, n)) for n in range(2, 12)}
    )

    return sets


class TestPyBitArray:
    def test_zeros(self) -> None:
        length = 50
        ba = bit_utils.PyBitArray(length, fill_bit=0)

        for idx in range(length):
            assert ba[idx] == 0

    def test_ones(self) -> None:
        length = 50
        ba = bit_utils.PyBitArray(length, fill_bit=1)

        for idx in range(length):
            assert ba[idx] == 1

    def test_set(self, sets: dict[str, set[int]]) -> None:
        universe = sets["universe"]
        size = len(universe)

        for name, test_set in sets.items():
            ba = bit_utils.PyBitArray(size, fill_bit=0)
            for p in test_set:
                ba[p] = 1

            for i in universe:
                if i in test_set:
                    assert ba[i] == 1, f"set: {name}; i: {i}"
                else:
                    assert ba[i] == 0, f"set: {name}; i: {i}"

    def test_unset(self, sets: dict[str, set[int]]) -> None:
        universe = sets["universe"]
        size = len(universe)

        for name, test_set in sets.items():
            ba = bit_utils.PyBitArray(size, fill_bit=1)

            complement = universe - test_set
            for c in complement:
                ba[c] = 0

            for i in universe:
                if i in test_set:
                    assert ba[i] == 1, f"set: {name}; i: {i}"
                else:
                    assert ba[i] == 0, f"set: {name}; i: {i}"

    def test_to_int(self, sets: dict[str, set[int]]) -> None:
        p100_int = 159085582874019712269820766380
        primes100 = sets["primes"]
        size = len(sets["universe"])
        ba = bit_utils.PyBitArray(size, fill_bit=0)
        for p in primes100:
            ba[p] = 1
        result = int(ba)
        assert result == p100_int

    def test_from_int(self, sets: dict[str, set[int]]) -> None:
        p100_int = 159085582874019712269820766380
        primes100 = sets["primes"]
        ba = bit_utils.PyBitArray.from_int(p100_int)

        for idx, bit in enumerate(ba):
            if bit:
                assert idx in primes100
            else:
                assert idx not in primes100


class TestOtherBits:
    def bits(self) -> None:
        vectors = [
            (0b1101, [1, 0, 1, 1]),
            (1, [1]),
            (0, []),
            (0o644, [0, 0, 1, 0, 0, 1, 0, 1, 1]),
            (65537, [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]),
        ]
        for n, expected in vectors:
            bits = [bit for bit in bit_utils.bits(n)]
            assert bits == expected

    def test_hamming(self) -> None:
        s1 = b"this is a test"
        s2 = b"wokka wokka!!!"

        hd = bit_utils.hamming_distance(s1, s2)
        assert hd == 37


class TestIntBits:
    """Tests for the bit manipulation of ints"""

    vectors: list[tuple[int, tuple[int, bool], int]] = [
        (0, (0, True), 1),
        (1, (0, True), 1),
        (1, (0, False), 0),
        (5, (1, True), 7),
        (5, (3, True), 13),
        (16, (4, False), 0),
        (16, (4, True), 16),
        (15, (2, False), 11),
        (15, (3, False), 7),
    ]

    def test_set_true(self) -> None:
        for left, (bit, value), right in self.vectors:
            if not value:
                left, right = right, left
            result = bit_utils.set_bit(left, bit, True)
            assert result == right

    def test_set_false(self) -> None:
        for left, (bit, value), right in self.vectors:
            if value:
                # left, right = right, left
                continue
            result = bit_utils.set_bit(left, bit, False)
            assert result == right

    def test_get_bit(self) -> None:
        vectors: list[tuple[str, dict[int, int]]] = [
            ("10101100", {0: 0, 1: 0, 2: 1, 3: 1, 4: 0}),
        ]
        for s, d in vectors:
            n = int(s, 2)
            for idx, expected in d.items():
                b = bit_utils.get_bit(n, idx)
                assert b == expected

    def test_index_linerar(self) -> None:
        vectors: list[tuple[int | str, dict[int, int | None]]] = [
            ("11010", {1: 1, 2: 3, 3: 4, 4: None}),
            (
                0b100010100010100010101100,
                {1: 2, 2: 3, 3: 5, 4: 7, 5: 11, 6: 13, 7: 17, 8: 19, 9: 23},
            ),
        ]

        for n, d in vectors:
            if isinstance(n, str):
                n = int(n, 2)
            for k, expected in d.items():
                result = bit_utils.bit_index_linear(n, k, 1)
                assert result == expected

    def test_index(self) -> None:
        vectors: list[tuple[int | str, dict[int, int | None]]] = [
            (
                pow(2, 300) - 1,
                {b: b - 1 for b in range(1, 300, 17)},
            ),
            ("11010", {1: 1, 2: 3, 3: 4, 4: None}),
            (
                0b100010100010100010101100,
                {1: 2, 2: 3, 3: 5, 4: 7, 5: 11, 6: 13, 7: 17, 8: 19, 9: 23},
            ),
        ]

        for n, d in vectors:
            if isinstance(n, str):
                n = int(n, 2)
            for k, expected in d.items():
                result = bit_utils.bit_index(n, k, 1)
                assert result == expected


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



from collections import UserDict
from random import sample
from typing import Any, Optional, TypeAlias
from itertools import combinations
from toy_crypto.bit_utils import hamming_distance

Letter: TypeAlias = str
"""Intended to indicate a str of length 1"""


class Alphabet:
    """An alphabet.

    This does not check if the alphabet is sensible. In particular, you may get
    very peculiar results if the alphabet contains duplicate elements.

    Instances of this class are conventionally immutable.
    """

    CAPS_ONLY = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    """'A' through 'Z' in order."""

    # Printable 7 bit ASCI with space but excluding backslash. Shuffled.
    PRINTABLE = r"""JDi-Km9247oBEctS%Isxz{<;=W^fL,[Y3Mgd6HV(kR8:_CF"*')>|#~Xay!]N+1vnqTl/}j$A.@0b ZGe`UPhp?Ow&ru5Q"""
    """
    Printable 7-bit ASCII in a fixed scrambled order.

    It does not include the backslash character,
    and the scrambled order is hardcoded.
     """

    DEFAULT = CAPS_ONLY
    """CAPS_ONLY is the default."""

    def __init__(
        self,
        alphabet: Optional[str] = None,
        prebaked: Optional[str] = None,
    ):
        """This does not check if the alphabet is sensible. In particular, you
        may get  very peculiar results if the alphabet contains duplicate
        elements.
        """

        match (alphabet, prebaked):
            case (None, None) | (None, "default"):
                abc = self.DEFAULT
            case (None, "caps"):
                abc = self.CAPS_ONLY
            case (None, "printable"):
                abc = self.PRINTABLE
            case (None, _):
                raise ValueError("Unknown pre-baked alphabet")
            case (_, None):
                if not isinstance(alphabet, str):
                    raise TypeError("alphabet must be a string")
                abc = alphabet
            case (_, _):
                raise ValueError(
                    "Can't use both explicit and pre-baked alphabet"
                )

        self._alphabet = abc

        self._modulus = len(self._alphabet)

        # Set up char to index table
        self._abc2idx: dict[Letter, int] = {
            c: i for i, c in enumerate(self._alphabet)
        }

    @property
    def alphabet(self) -> str:
        """The underlying alphabet."""
        return self._alphabet

    @property
    def modulus(self) -> int:
        """The modulus."""
        return self._modulus

    @property
    def abc2idx(self) -> dict[Letter, int]:
        """Dictionary of letter to position in the alphabet."""
        return self._abc2idx

    # We will want to use 'in' for Alphabet instances
    def __contains__(self, item: Any) -> bool:
        """
        Allows the 'in' and 'not in' operators.

        So if `abc` is an Alphabet, ``'Z' in abc`` is well defined.
        """
        return item in self.alphabet

    def __getitem__(self, index: slice | int) -> str:
        """Allows retrieving bits of the Alphabet through [index] notation."""
        return self.alphabet[index]

    # annoyingly, the type str is also used for single character strings
    # add, inverse, subtract all deal with single characters
    def add(self, a: Letter, b: Letter) -> Letter:
        """Returns the modular sum of two characters."""
        if a not in self or b not in self:
            raise ValueError("argument not an element")
        idx = (self.abc2idx[a] + self.abc2idx[b]) % self.modulus
        return self.alphabet[idx]

    def inverse(self, c: Letter) -> Letter:
        """Returns the additive inverse of character c"""
        if c not in self:
            raise ValueError("argument not an element")
        idx = (self.modulus - self.abc2idx[c]) % self.modulus
        return self.alphabet[idx]

    def subtract(self, a: Letter, b: Letter) -> Letter:
        """Returns the character corresponding to a - b."""
        return self.add(a, self.inverse(b))


class Cipher:
    """A Vigenère Cipher is a key and an alphabet."""

    def __init__(self, key: str, alphabet: Alphabet | str | None = None):
        if isinstance(alphabet, Alphabet):
            abc = alphabet
        else:
            abc = Alphabet(alphabet)

        self._alphabet = abc

        if not key:
            raise ValueError("key must not be empty")

        if any([k not in self._alphabet for k in key]):
            raise ValueError(
                "key must be comprised of characters in the alphabet"
            )
        self._key: str = key
        self._key_length = len(self._key)

    @property
    def alphabet(self) -> Alphabet:
        """The Alphabet for this cipher."""
        return self._alphabet

    @property
    def key(self) -> str:
        """Shhh! This is the key. Keep it secret."""
        return self._key

    @property
    def modulus(self) -> int:
        """The modulus."""
        return self._alphabet.modulus

    def crypt(self, text: str, mode: str) -> str:
        """{en,de}crypts text depending on mode"""

        match mode:
            case "encrypt":
                operation = self.alphabet.add
            case "decrypt":
                operation = self.alphabet.subtract
            case _:
                raise ValueError("mode must be 'encrypt' or 'decrypt")

        # TODO: Generalize this for streaming input and output
        output: list[Letter] = []

        """
        I would love to use zip and cycle, but I need to handle input
        characters that are not in the alphabet.
        """

        key_idx = 0
        for c in text:
            if c not in self.alphabet:
                result = c
            else:
                k = self.key[key_idx]
                result = operation(c, k)
                key_idx = (key_idx + 1) % self._key_length

            output.append(result)

        return "".join(output)

    def encrypt(self, plaintext: str) -> str:
        """Returns ciphertext."""

        return self.crypt(plaintext, mode="encrypt")

    def decrypt(self, ciphertext: str) -> str:
        """Returns plaintext."""

        return self.crypt(ciphertext, mode="decrypt")


BitSimilarity = float
"""A float in [-4.0, 4.0] the bit similarity per byte.
-4 indicates that all bits are different.
+4 indicates that all bits are the same.
"""


class SimilarityScores(UserDict[int, list[BitSimilarity]]):
    """A dictionary of keysize : list[BitSimilarity]."""

    def __init__(self) -> None:
        self.data: dict[int, list[BitSimilarity]] = {}
        self._best: Optional[int] = None

    @staticmethod
    def is_keysize(k: object) -> bool:
        if not isinstance(k, int):
            return False
        return k >= 0

    @staticmethod
    def is_bit_similarity(x: object) -> bool:
        if not isinstance(x, float | int):
            return False
        return x >= -4.0 and x <= 4.0

    def is_valid(self) -> bool:
        if not all(SimilarityScores.is_keysize(k) for k in self.data.keys()):
            return False

        for scores in self.data.values():
            if not all(SimilarityScores.is_bit_similarity(s) for s in scores):
                return False
        return True

    def mean(self, k: int) -> float:
        """The average score for keysize k.

        :raises KeyError: if k triggers a KeyError.
        """

        scores = self.data[k]
        return sum(scores) / len(scores)

    @property
    def best(self) -> int:
        """Keysize with the best (highest average) score."""

        if self._best is not None:
            return self._best

        best_so_far: tuple[int, float] = (0, -4.0)

        for k in self.data.keys():
            mean = self.mean(k)
            if mean > best_so_far[1]:
                best_so_far = (k, mean)

        self._best = best_so_far[0]
        return self._best


def probable_keysize(
    ciphertext: bytes | str,
    min_size: int = 3,
    max_size: int = 40,
    trial_pairs: int = 1,
) -> SimilarityScores:
    """Assesses likelihood for key length of ciphertext.

    :param ciphertext: The ciphertext.
    :param min_size: The minimum key length to try.
    :param max_size: The maximum key length to try.
    :param trial_pairs: The number of pairs of blocks to test.

    :return: Returns list sorted by scores of (keysize, score)

    Scores are scaled 0 (least likely) to 1 (most likely),
    but they do not directly represent probabilities.
    """

    scores = SimilarityScores()

    if min_size == max_size:
        # Should this be a ValueError?
        scores[min_size] = [0]
        return scores

    if min_size > max_size:
        raise ValueError("min_size can't be larger than max_size")

    if trial_pairs < 1:
        raise ValueError("trial_pairs must be positive")

    if isinstance(ciphertext, str):
        ciphertext = bytes(ciphertext, encoding="utf8")

    ctext_len = len(ciphertext)
    for keysize in range(min_size, max_size):
        if 2 * keysize > ctext_len:
            continue
        num_blocks = ctext_len // keysize
        all_pairs = list(combinations(range(num_blocks), 2))

        # trial_pairs may have to be reduced to
        trial_pairs = min([trial_pairs, len(all_pairs)])

        pairs = sample(all_pairs, trial_pairs)

        def get_block(idx: int) -> bytes:
            idx *= keysize
            return ciphertext[idx : idx + keysize]

        s_scores: list[float] = []
        for i, j in pairs:
            distance = hamming_distance(get_block(i), get_block(j))
            similarity_per_byte = 4.0 - (distance / keysize)
            s_scores.append(similarity_per_byte)

        scores[keysize] = s_scores

    return scores



# SPDX-FileCopyrightText: 2024-present Jeffrey Goldberg <jeffrey@goldmark.org>
#
# SPDX-License-Identifier: MIT

import os
import sys

PROJECT_PATH = os.getcwd()
SOURCE_PATH = os.path.join(PROJECT_PATH, "src")
sys.path.append(SOURCE_PATH)



"""Simple elliptic curve groups."""

import sys

if sys.version_info < (3, 11):
    raise Exception("Requires python 3.11")
from typing import Optional, Self
from functools import cached_property

"""
Until some of the self-type issues with mypy are better sorted out,
use of Self is minimal.

https://github.com/python/mypy/labels/topic-self-types
"""

from toy_crypto.nt import Modulus as Modulus  # noqa: E402
from toy_crypto.nt import is_modulus, mod_sqrt  # noqa: E402
from toy_crypto.bit_utils import bits  # noqa: E402


"""
This is more complicated because I unwisely attempted to make this work both
for elliptic curves defined over the reals (for the drawings) and for curves
defined over integer fields.
"""


class Curve:
    def __init__(self, a: int, b: int, p: int) -> None:
        """Define a curve of the form :math:`y^2 = x^3 + ax + b \\pmod p`."""
        self._p: Modulus = Modulus(p)
        self._a: int = a
        self._b: int = b

        if self.is_singular():
            raise ValueError(f"{self} is singular")

        if not is_modulus(self.p):
            raise ValueError("Bad modulus p")

        self._pai: Optional[Point] = None

        # This assumes (without checking) that the curve has good parameters
        # and that a generator (base point) has been chosen correctly/
        self._order = (self.p + 1) // 2

    @property
    def a(self) -> int:
        """The 'a' of :math:`y^2 = x^3 + ax + b \\pmod p`."""
        return self._a

    @property
    def b(self) -> int:
        """The 'b' of :math:`y^2 = x^3 + ax + b \\pmod p`."""
        return self._b

    @property
    def p(self) -> Modulus:
        """The 'p' of :math:`y^2 = x^3 + ax + b \\pmod p`."""
        return self._p

    def is_singular(self) -> bool:
        return (4 * self._a**3 + 27 * self._b * self._b) % self._p == 0

    @cached_property
    def PAI(self) -> "Point":
        """Point At Infinity: the additive identity."""
        if self._pai is None:
            self._pai = Point(0, 0, self)
            self._pai._is_mutable = False
            self._pai._is_pai = True
        return self._pai

    @property
    def order(self) -> int:
        """The order of the group."""
        return self._order

    def __repr__(self) -> str:
        # There is probably a nice way to do with with
        # format directives, but I'm not going to dig
        # into those docs now.
        if self._a < 0:
            ax = f"- {-self._a}x"
        else:
            ax = f"+ {self._a}"
        if self.b < 0:
            b = f"- {-self._b}x"
        else:
            b = f"+ {self.b}"

        return f"y^2 = x^3 {ax} {b} (mod {self._p})"

    def compute_y(self, x: int) -> Optional[tuple[int, int]]:
        "Returns pair of y values for x on curve. None otherwise."
        a = self._a
        b = self._b
        p = self._p
        y2: int = (pow(x, 3, p) + ((a * x) % p) + b) % p
        roots = mod_sqrt(y2, p)
        if len(roots) != 2:
            raise ValueError("x is rootless")

        return roots[0], roots[1]

    def point(self, x: int, y: int) -> "Point":
        return Point(x, y, self)


class Point:
    """Point on elliptic curve over finite field."""

    def __init__(self, x: int, y: int, curve: Curve) -> None:
        """Create a mutable point on a curve."""

        self._x: int = x
        self._y: int = y
        self._curve = curve

        # We are creating a normal point, so we set these
        self._is_mutable: bool = True
        self._is_pai = False

        if not (isinstance(self._x, int) and isinstance(self._y, int)):
            raise TypeError("Points must have integer coordinates")

        self._x %= self._curve.p
        self._y %= self._curve.p

        if not self.on_curve():
            raise ValueError("point not on curve")

    @property
    def x(self) -> int:
        return self._x

    @property
    def y(self) -> int:
        return self._y

    @property
    def curve(self) -> Curve:
        return self._curve

    @property
    def is_zero(self) -> bool:
        """True if point at infinity"""
        return self._is_pai

    def on_curve(self) -> bool:
        """True if point is on the curve (including point at infinity)."""
        if self._is_pai:
            return True

        x = int(self._x)
        y = int(self._y)
        a = int(self.curve.a)
        b = int(self.curve.b)

        p = self._curve.p
        # breaking this down for debugging
        lhs = pow(y, 2, p)
        rhs = (pow(x, 3, p) + a * x + b) % p

        return lhs == rhs

    # define P + Q; -P; P += Q;  P - Q; P == Q
    def __add__(self: Self, Q: Self) -> Self:
        return self.add(Q)

    def __neg__(self) -> Self:
        return self.neg()

    def __iadd__(self: Self, Q: Self) -> Self:
        return self.iadd(Q)

    def __sub__(self, Q: Self) -> Self:
        return self.__add__(Q.__neg__())

    def __eq__(self, Q: object) -> bool:
        if not isinstance(Q, Point):
            return NotImplemented
        if not self and not Q:  # both are 0
            return True
        if self._x != Q.x or self._y != Q.y:  # x's and y's don't match
            return False
        if self._curve != Q.curve:  # They are defined for different curves
            return False
        return True

    def __bool__(self) -> bool:
        """P is True iff P is not the zero point."""
        return not self._is_pai

    def __repr__(self) -> str:
        return f"({self.x}, {self.y})"

    def neg(self: Self) -> Self:
        """Return additive inverse.

        :returns: Additive inverse
        :rtype: Point
        """

        if self.is_zero:
            return self

        r = self.cp()
        r._y = self._curve.p - r._y
        return r

    # I don't know how shallow a copy() is in Python, so
    def cp(self: Self) -> Self:
        """Return a mutable copy"""

        c = type(self)(self._x, self._y, self._curve)
        c._is_pai = self._is_pai
        c._is_mutable = True
        return c

    def iadd(self: Self, Q: Self) -> Self:
        """add point to self in place.

        The Point at Infinity is not mutable.
        But you can make a mutable copy
        with ``curve.PAI.cp()``.
        Or you could just use :meth:`Point.add`.

        :raises ValueError: if Q is not on its own curve
        :raises ValueError: if Q and self are on different curves
        :raises NotImplementedError: if self is not a mutable point.
        """

        if not self._is_mutable:
            raise NotImplementedError("Point must be mutable")

        # The order of checking matters, as each check is seen as
        # as a fall through of prior checks
        if not isinstance(Q, Point):
            return NotImplemented

        # We don't do curve check on Q if Q is 0
        # P + 0 = P
        if not Q:
            return self

        # if Q is not on its curve then there is something wrong with it.
        if not Q.on_curve():
            raise ValueError("Point is not on curve")

        # 0 + Q = Q
        if not self:
            self._x, self._y = Q.x, Q.y
            self._curve = Q.curve
            self._is_pai = Q.is_zero
            return self

        # if Q is on a different curve, something bad is happening
        if self.curve != Q.curve:
            raise ValueError("Points on different curves")

        # P + P
        if self == Q:
            return self.idouble()

        # P + -P = 0
        if self._x == Q.x:
            self._x, self._y = 0, 0
            self._is_pai = True
            return self

        # Generics would be better than the abuse of type
        # narrowing that I am doing here to call different
        # _addition() methods

        self._x, self._y = self._nz_addition(Q)

        return self

    def add(self: Self, Q: Self) -> Self:
        """Add points.

        :param Q: Point to add
        :type Q: Point

        :returns: Sum of Q and self
        :rtype: Point
        """

        r = self.cp()
        r.iadd(Q)
        return r

    def idouble(self) -> Self:
        """Double point in place.

        The Point at Infinity is not mutable.
        But you can make a mutable copy
        with ``curve.PAI.cp()``
        or just use :meth:`Point.double`.

        :raises NotImplementedError: if self is not a mutable point.
        """

        if not self._is_mutable:
            raise NotImplementedError(
                "Point must be mutable for in-place operation."
            )
        if self.is_zero:
            return self

        xy = self._xy_double()
        if not xy:
            self._x = 0
            self._y = 0
            self._is_pai = True
        else:
            self._x, self._y = xy

        return self

    def double(self: Self) -> Self:
        """Returns self + self"""
        if self.is_zero:
            return self.cp()

        P = self.cp()
        P = P.idouble()

        return P

    def _xy_double(self) -> Optional[tuple[int, int]]:
        """(x, y) for x, y of doubled point. None if point at infinity

        :returns: new coordinates, x and y
        :rtype: Optional[tuple[int,int]]


        This does _some_ validity check of input values,
        but it might just return erroneous results the following
        conditions aren't met
        - self.curve.p is prime
        - self.curve is well defined
        - self.x, self.y are integers
        - self is on the curve
        - self is not the point at infinity
        """

        if self._is_pai:
            return None

        if self._y == 0:
            return None

        m = self._curve.p
        top = ((3 * (self._x * self._x)) % m + self._curve.a) % m
        bottom = (2 * self.y) % m
        inv_bottom = pow(bottom, m - 2, m)
        s = top * inv_bottom % m

        x = (pow(s, 2, m) - 2 * self.x) % m
        y = (s * (self.x - x) - self.y) % m

        return (x, y)

    def scaler_multiply(self, n: int) -> "Point":
        """
        Returns n * self.

        :warning: This algorithm exposes n to timing attacks.
        """

        n = n % self.curve.order
        sum = self.curve.PAI.cp()  # additive identity
        doubled = self
        for bit in bits(n):
            if bit == 1:
                sum += doubled
            doubled = doubled.double()  # toil and trouble
        return sum

    def _nz_addition(self, Q: Self) -> tuple[int, int]:
        """returns x, y over finite field Z_p"""

        if self._is_pai or Q.is_zero:
            raise ValueError("this is for non-zero points only")

        m = self.curve.p  # have the modulus handy

        # The following breaks up the point addition math into
        # gory details and steps. This helped in debugging.

        # s = (Q.y - self.y) / (Q.x - self.x)
        #   = top/bot
        #   = top * inv_bot
        #
        # And we do our mod p reductions at every opportunity
        bottom = Q.x - self.x
        # because p is prime, we can use a^{p-2} % p to compute inverse of a
        inv_bot = pow(bottom, m - 2, m)
        top = (Q.y - self.y) % m
        s = top * inv_bot % m
        s2 = (s * s) % m

        # x = (s^2 - Px) - Qx
        x = (s2 - self.x) - Q.x
        x %= m

        # y = s(Px - x) - Qy
        y = s * (self.x - x) - self.y
        y %= m

        return x, y



import io
import polars as pl
from polars.datatypes import UInt32, Float32
from polars.dataframe.frame import DataFrame as PolarsDF
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.figure import Figure


def load_data() -> PolarsDF:
    """Idiosyncratic things for building our specific dataframe."""

    cvs_data = b"""
        size,Sieve,IntSieve,SetSieve
        100,0.00017137499526143074,0.000017999671399593353,0.000042750034481287
        1000,0.000016292091459035873,0.00017720786854624748,0.0000866670161485672
        10000,0.00010245898738503456,0.007494000252336264,0.0008742911741137505
        100000,0.0010270839557051659,0.5487742917612195,0.009959292132407427
        1000000,0.010919792111963034,59.795512250158936,0.19800291722640395
        50,2.0833685994148254e-6,0.000013249926269054413,0.000014875084161758423
        500,1.2922100722789764e-6,0.00007804157212376595,0.00004054093733429909
        5000,1.25030055642128e-6,0.002527667209506035,0.0004529156722128391
        50000,1.3331882655620575e-6,0.14295708294957876,0.004889749921858311
        500000,7.2498805820941925e-6,15.271088582929224,0.07385779218748212
    """.replace(b" ", b"")

    # We don't need 64 bit precision
    schema = {
        "size": UInt32,
        "bitarray": Float32,
        "int": Float32,
        "set": Float32,
    }

    CVS_DATA = io.BytesIO(cvs_data)
    # CVS_FILE = "timings.csv"
    df_wide = pl.read_csv(
        CVS_DATA,
        schema=schema,
        # skip_lines=1,
        # has_header=False,
    )

    df = df_wide.unpivot(
        index="size", variable_name="sieve_type", value_name="time"
    )
    return df


def base_g(data: PolarsDF, title: str | None = None) -> Figure:
    """Returns a FacetGrid which can be the base for several plots.

    For reasons I don't understand, it appears that once a plot is shown
    or saved, the object is destroyed. And I didn't find a way create
    a copy of a plot or subplot in a usable way.
    """

    # We want to keep colors constant even for graph that
    # doesn't have "py_int"
    hue_order = ["bitarray", "set"]
    if "int" in data["sieve_type"]:
        hue_order.append("int")

    g = sns.relplot(
        data=data,
        kind="line",
        x="size",
        y="time",
        hue="sieve_type",
        hue_order=hue_order,
        linewidth=2,
    )
    g.set(
        xscale="log",
        xlabel="Sieve size",
        ylabel="Time (seconds)",
    )
    # This does not feel cool, but is the best I can do for now
    g._legend.set_title("Sieve type")  # type: ignore

    if title:
        g.figure.subplots_adjust(top=0.95)
        g.set(title=title)

    return g.figure


def main() -> None:
    df = load_data()

    sns.set_theme(
        style="white",
        palette="colorblind6",
        font="Fira Sans",
    )

    # Whole plot
    fig = base_g(df, title="Sieve creation timings for size up to 1 million")
    fig.savefig("all_data.png")
    plt.show()

    xmax = 10**4
    df_e4 = df.filter(pl.col("size") <= xmax)
    fig = base_g(df_e4, title="Sieve creation timings for size up to 10000")
    fig.savefig("to_10_000.png")
    plt.show()

    df_sans_int = df.filter(pl.col("sieve_type") != "int")
    fig = base_g(
        df_sans_int, title="Sieve creation times for bitarray and set only"
    )
    fig.savefig("sans_int.png")
    plt.show()


if __name__ == "__main__":
    main()



from functools import cache
from bisect import bisect_right
import threading
from typing import (
    Any,
    Iterator,
    Protocol,
    Self,
    Union,
    runtime_checkable,
    TYPE_CHECKING,
)

from . import bit_utils
from math import isqrt


_has_bitarry = True
if TYPE_CHECKING:
    from bitarray import bitarray
    from bitarray.util import count_n, ba2int

else:
    try:
        from bitarray import bitarray
        from bitarray.util import count_n, ba2int
    except ImportError:
        _has_bitarry = False

        def bitarray(*args, **kwargs) -> Any:  # type: ignore
            raise NotImplementedError("bitarray is not installed")

        def count_n(*args, **kwargs) -> int:  # type: ignore
            raise NotImplementedError("bitarray is not installed")

        def ba2int(*args, **kwargs) -> int:  # type: ignore
            raise NotImplementedError("bitarray is not installed")


@runtime_checkable
class Sievish(Protocol):
    """Methods available for all Sieve-like classes.

    This is primary of use for testing, where one might need to write
    functions that interact with any of the Sieve classes.
    This also would probably make more sense as an abstract class
    instead of a Protocol.
    """

    @classmethod
    def _reset(cls) -> None:
        """Resets the class largest sieve created if such a thing exists.

        This exists only for running tests on caching.

        For classes that internally share data among instances, this is likely
        to leave instances in an inconsistent state.
        This is a no-op for classes that do not cache the largest sieve they
        created.
        """
        ...

    _data: Union[bitarray, list[int], int]

    @property
    def count(self) -> int:
        """The total number of primes in the sieve"""
        ...

    @property
    def n(self) -> int:
        """The size of the sieve, including composites.

        The number n such that the sieve contains all primes <= n.
        """
        ...

    def primes(self, start: int = 1) -> Iterator[int]:
        """Iterator of primes starting at start-th prime.

        The 1st prime is 2. There is no zeroth prime.

        :raises ValueError: if start < 1
        """
        ...

    def nth_prime(self, n: int) -> int:
        """Returns n-th prime.

        :raises ValueError: if n exceeds count.
        :raises ValueError: n < 1
        """
        ...

    def __int__(self) -> int:
        """Sieve as an integer with 1 bits representing primes.

        Most significant 1 bit represents the largest prime in the sieve.
        For example if s is a sieve of size 5, ``int(s)`` returns 44 which
        is equivalent to 0b101100.
        """
        ...

    @classmethod
    def from_size[S](cls: type[S], size: int) -> S:
        """Returns a new sieve of primes less than or equal to size."""
        ...

    @classmethod
    def from_int[S](cls: type[S], n: int) -> S:
        """Returns a new sieve of primes from the bits of n."""
        ...

    @classmethod
    def from_list[S](
        cls: type[S], primes: list[int], size: int | None = None
    ) -> S:
        """Returns a new sieve of primes from list.

        If size is not specified it will be set to the largest value in primes

        :raises ValueError: if primes is empty
        """
        ...


class BaSieve(Sievish):
    """Sieve of Eratosthenes.

    The good parts of this implementation are lifted from the example provided
    with the `bitarray package <https://pypi.org/project/bitarray/>`_ source.

    This depends on `bitarray package <https://pypi.org/project/bitarray/>`_.
    """

    _base_sieve = bitarray("0011")

    _common_data = bitarray("0011")
    _largest_n: int = 2
    """These will be shared by all instances."""

    lock = threading.Lock()

    @classmethod
    def _extend(cls, n: int) -> None:
        if n <= cls._largest_n:
            return

        with cls.lock:
            len_e = n - cls._largest_n
            cls._common_data.extend([True] * len_e)

            """start is the multiple of the prime we start zeroing the
            array from. Typically that would be 2p, but we want to consider
            cases where the common_data is larger than n. All composites
            through largest largest_n have already been set to 0 in the array.

            So ``start`` must meet four conditions
            1. It must be a multiple of p
            2. It must be larger than p
            3. It must be larger than largest_n
            4. It must be the smallest number meeting the above conditions
            """

            start: int
            for p in range(2, isqrt(n) + 1):
                if cls._common_data[p]:
                    start = max(p + p, p * (p // cls._largest_n + 1))
                    cls._common_data[start::p] = 0
            cls._largest_n = n

    @classmethod
    def _reset(cls) -> None:
        with cls.lock:
            cls._common_data = cls._base_sieve.copy()
            cls._largest_n = 2

    @classmethod
    def from_int(cls, n: int) -> Self:
        instance = cls.__new__(cls)
        instance._n = n.bit_length()
        instance._data = bitarray(instance._n + 1)
        idx = 0
        while n:
            n, r = divmod(n, 2)
            instance._data[idx] = r
            idx += 1

        instance._count = instance._data.count()

        return instance

    @classmethod
    def from_size(cls, size: int) -> Self:
        if size < 2:
            raise ValueError("size must be greater than 2")

        instance = cls.__new__(cls)

        instance._extend(size)
        instance._n = size

        instance._count = instance._common_data[:size].count()

        return instance

    @classmethod
    def from_list(cls, primes: list[int], size: int | None = None) -> Self:
        if not primes:
            raise ValueError("primes cannot be empty")
        primes = sorted(primes)
        max_prime = primes[-1]
        assert isinstance(max_prime, int)

        if size is None:
            size = max_prime

        instance = cls.__new__(cls)
        if max_prime > cls._largest_n:
            with cls.lock:
                extend_by = max_prime - cls._largest_n
                cls._common_data.extend([0] * extend_by)

                for p in primes:
                    if p > cls._largest_n:
                        instance._common_data[p] = 1
                cls._largest_n = max(size, max_prime)
        instance._n = size
        instance._count = instance._common_data[:size].count()

        return instance

    def __init__(self, data: bitarray, size: int | None) -> None:
        """Sieve from bitarray, treated as up to size.

        :raises ValueError: if size > ``len(data)``
        """

        if size is None:
            size = len(data)

        if size > len(data):
            raise ValueError(
                "size cannot be larger than the length of the data"
            )
        if size > self._largest_n:
            with self.lock:
                self._largest_n = len(data)
                self._common_data = data

        self._n = size

        self._count: int = self._common_data[: self._n].count()

    @property
    def n(self) -> int:
        return self._n

    @property
    def count(self) -> int:
        """The number of primes in the sieve."""
        return self._count

    @cache
    def nth_prime(self, n: int) -> int:  # pyright: ignore
        assert isinstance(self._common_data, bitarray)
        if n < 1:
            raise ValueError("n must be greater than zero")

        if n > self._count:
            raise ValueError("n cannot exceed count")

        return count_n(self._common_data, n)

    def primes(self, start: int = 1) -> Iterator[int]:
        if start < 1:
            raise ValueError("Start must be >= 1")
        for n in range(start, self._count + 1):
            yield count_n(self._common_data, n) - 1

    def __int__(self) -> int:
        reversed = self._common_data.copy()[: self._n]
        reversed.reverse()
        return ba2int(reversed)

    # "Inherit" docstrings. Can't do properties
    from_size.__doc__ = Sievish.from_size.__doc__
    __int__.__doc__ = Sievish.__int__.__doc__
    primes.__doc__ = Sievish.primes.__doc__
    _reset.__doc__ = Sievish._reset.__doc__
    nth_prime.__doc__ = Sievish.nth_prime.__doc__
    from_int.__doc__ = Sievish.from_int.__doc__
    from_list.__doc__ = Sievish.from_list.__doc__


class SetSieve(Sievish):
    """Sieve of Eratosthenes using a native python set

    This consumes an enormous amount of early in initialization,
    and a SetSieve object will contain a list of prime integers,
    so even after initialization is requires more memory than the
    the integer or bitarray sieves.
    """

    _base_sieve: list[int] = [2, 3]
    _common_data = _base_sieve.copy()
    _largest_n: int = 3

    lock = threading.Lock()

    @classmethod
    def _extend(cls, n: int) -> None:
        if n <= cls._largest_n:
            return

        with cls.lock:
            largest_p = cls._common_data[-1]
            upper_set: set[int] = {c for c in range(largest_p + 1, n + 1)}

            # first we sieve out products of the primes we
            # already have from the upper set.
            #
            # See comments about ``start`` in BaSieve
            start: int
            for p in range(2, isqrt(n) + 1):
                if p in cls._common_data:
                    start = max(p + p, p * (p // cls._largest_n + 1))
                    p_products = range(start, n + 1, p)
                    for c in p_products:
                        upper_set.discard(c)

            # Now we sieve out products of things remaining in the upper set
            for p in range(largest_p, isqrt(n) + 1):
                if p in upper_set:
                    for c in range(2 * p, n + 1, p):
                        upper_set.discard(c)

            cls._common_data.extend(sorted(upper_set))
            cls._largest_n = n

    @classmethod
    def _reset(cls) -> None:
        pass

    @classmethod
    def from_int(cls, n: int) -> Self:
        instance = cls.__new__(cls)
        instance._n = n.bit_length()
        if n.bit_length() > cls._largest_n:
            with cls.lock:
                new_primes = (
                    p
                    for p, b in enumerate(bit_utils.bits(n))
                    if p > cls._largest_n and b
                )
                cls._common_data.extend(new_primes)
                cls._largest_n = n.bit_length()

        return instance

    @classmethod
    def from_list(cls, primes: list[int], size: int | None = None) -> Self:
        if not primes:
            raise ValueError("primes cannot be empty")

        # not only sorts, but gives us a copy
        primes = sorted(primes)

        if size is None:
            size = primes[-1]
        elif size < 2:
            raise ValueError("size must be greater than 2")

        instance = cls.__new__(cls)
        instance._n = size
        if len(primes) > len(cls._common_data):
            with cls.lock:
                ip = bisect_right(primes, cls._largest_n)
                new_primes = primes[ip:]
                cls._common_data.extend(new_primes)
                cls._largest_n = size

        return instance

    def __init__(self, data: list[int], size: int | None = None) -> None:
        """Returns sorted list primes n =< n

        A pure Python (memory hogging) Sieve of Eratosthenes.
        This consumes lots of memory, and is here only to
        illustrate the algorithm.
        """

        # not only sorts, but gives us a copy
        data = sorted(data)

        if size is None:
            size = data[-1]

        with self.lock:
            if size > self._largest_n:
                self._largest_n = size
                self._common_data = data
        self._n = size

    @classmethod
    def from_size[S](cls, size: int) -> "SetSieve":
        if size < 2:
            raise ValueError("size must be greater than 2")

        instance = cls.__new__(cls)

        instance._n = size
        instance._extend(size)

        return instance

    @property
    def count(self) -> int:
        if self._n < self._largest_n:
            ip = bisect_right(self._common_data, self._n)
            return len(self._common_data[:ip])
        return len(self._common_data)

    def primes(self, start: int = 1) -> Iterator[int]:
        if start < 1:
            raise ValueError("Start must be >= 1")

        for n in range(start, self.count + 1):
            yield self._common_data[n - 1]

    def nth_prime(self, n: int) -> int:
        """Returns n-th prime. ``nth_prime(1) == 2``. There is no zeroth prime.

        :raises ValueError: if n exceeds count.
        :raises ValueError: n < 1
        """

        if n < 1:
            raise ValueError("n must be greater than zero")

        if n > self.count:
            raise ValueError("n cannot exceed count")

        return self._common_data[n - 1]

    def __int__(self) -> int:
        ip = bisect_right(self._common_data, self._n)
        result = sum((int(2**p) for p in self._common_data[:ip]))
        return result

    @property
    def n(self) -> int:
        return self._n

    from_size.__doc__ = Sievish.from_size.__doc__
    __int__.__doc__ = Sievish.__int__.__doc__
    primes.__doc__ = Sievish.primes.__doc__
    _reset.__doc__ = Sievish._reset.__doc__
    nth_prime.__doc__ = Sievish.nth_prime.__doc__
    from_int.__doc__ = Sievish.from_int.__doc__
    from_list.__doc__ = Sievish.from_list.__doc__


class IntSieve(Sievish):
    """A pure Python (using a large int) Sieve of Eratosthenes."""

    _BASE_SIEVE: int = int("1100", 2)

    @classmethod
    def _reset(cls) -> None:
        pass

    @classmethod
    def from_int(cls, n: int) -> Self:
        sieve = cls.__new__(cls)
        sieve._n = n.bit_length()
        sieve._data = n
        sieve._count = sieve._data.bit_count()

        return sieve

    @classmethod
    def from_list(cls, primes: list[int], size: int | None = None) -> Self:
        if not primes:
            raise ValueError("primes cannot be empty")
        max_prime = max(primes)
        assert isinstance(max_prime, int)
        if size is None:
            size = max_prime

        instance = cls.__new__(cls)
        instance._data = sum((int(2**p) for p in primes))
        instance._n = size
        instance._count = instance._data.bit_count()
        return instance

    def __init__(self, data: int) -> None:
        self._data: int = data
        self._n = self._data.bit_length()
        self._count = self._data.bit_count()

    def _extend(self, n: int) -> None:
        if n <= self._data.bit_length():
            return
        ones = (2 ** ((n - self._n) + 1)) - 1
        ones = ones << self._n
        self._data |= ones

        self._n = n
        # We only need to go up to and including the square root of n,
        # remove all non-primes above that square-root =< n.
        for p in range(2, isqrt(n) + 1):
            # if utils.get_bit(self._data, p):
            if (self._data & (1 << p)) >> p:
                # Because we are going through sieve in numeric order
                # we know that multiples of anything less than p have
                # already been removed, so p is prime.
                # Our job is to now remove multiples of p
                # higher up in the sieve.
                for m in range(p + p, n + 1, p):
                    # self._data = utils.set_bit(self._data, m, False)
                    self._data = self._data & ~(1 << m)  # pyright: ignore

    @classmethod
    def from_size[S](cls, size: int) -> "IntSieve":
        if size < 2:
            raise ValueError("size must be greater than 2")

        instance = cls.__new__(cls)

        instance._data = instance._BASE_SIEVE
        instance._n = instance._BASE_SIEVE.bit_length()
        instance._extend(size)
        instance._count = instance._data.bit_count()
        return instance

    def nth_prime(self, n: int) -> int:
        if n < 1:
            raise ValueError("n must be greater than zero")

        if n > self.count:
            raise ValueError("n cannot exceed count")

        result = bit_utils.bit_index(self._data, n)
        assert result is not None  # because we checked n earlier
        return result

    @property
    def count(self) -> int:
        return self._count

    @property
    def n(self) -> int:
        return self._n

    def primes(self, start: int = 1) -> Iterator[int]:
        if start < 1:
            raise ValueError("Start must be >= 1")
        for n in range(start, self.count + 1):
            pm = bit_utils.bit_index(self._data, n)
            assert pm is not None
            yield pm

    def __int__(self) -> int:
        return self._data

    # 'Inherit' docstrings
    from_size.__doc__ = Sievish.from_size.__doc__
    __int__.__doc__ = Sievish.__int__.__doc__
    primes.__doc__ = Sievish.primes.__doc__
    _reset.__doc__ = Sievish._reset.__doc__
    nth_prime.__doc__ = Sievish.nth_prime.__doc__
    from_int.__doc__ = Sievish.from_int.__doc__
    from_list.__doc__ = Sievish.from_list.__doc__


# https://mypy.readthedocs.io/en/stable/common_issues.html#variables-vs-type-aliases
Sieve: type[object]
"""Sieve will be an alias for BaSieve if bitarray is available,
otherwise it will be assigned to some other sieve class."""

if _has_bitarry:
    Sieve = BaSieve
else:
    Sieve = SetSieve



from collections import namedtuple
from typing import Optional


from toy_crypto import rsa
from toy_crypto.nt import lcm, modinv
from toy_crypto.utils import Rsa129


class TestCitm:
    """Tests using Cat in the Middle story"""

    # Not really a great set of tests, since the data I am testing against
    # was created with some of the same code I'm testing

    e = 17

    class Critter:
        def __init__(
            self,
            factors: tuple[int, int],
            expected_N: int,
            expected_d: Optional[int] = None,
        ) -> None:
            self.factors = factors
            self.expected_N = expected_N
            self.expected_d = expected_d

            self.test_data: list[tuple[int, int]] = []

    e = 17

    patty = Critter((107, 151), 16157, 1403)
    patty.test_data = [(1234, 8900)]

    molly = Critter((97, 43), 4171, 593)
    molly.test_data = [(1313, 530), (1729, 2826)]

    mr_talk = Critter((47, 89), 4183, 1905)
    mr_talk.test_data = [(1729, 2016)]

    def test_encrypt(self) -> None:
        for critter in [self.patty, self.molly, self.mr_talk]:
            p, q = critter.factors
            key = rsa.PrivateKey(p, q, pub_exponent=self.e)
            pubkey = key.pub_key

            for ptext, ctext in critter.test_data:
                assert ctext == pubkey.encrypt(ptext)

    def test_decrypt(self) -> None:
        for critter in [self.patty, self.molly, self.mr_talk]:
            p, q = critter.factors
            key = rsa.PrivateKey(p, q, pub_exponent=self.e)

            for ptext, ctext in critter.test_data:
                assert ptext == key.decrypt(ctext)

    def test_N(self) -> None:
        for critter in [self.patty, self.molly, self.mr_talk]:
            p, q = critter.factors
            key = rsa.PrivateKey(p, q, pub_exponent=self.e)
            pubkey = key.pub_key

            assert critter.expected_N == pubkey.N

    def test_d(self) -> None:
        for critter in [self.patty, self.molly, self.mr_talk]:
            p, q = critter.factors
            key = rsa.PrivateKey(p, q, pub_exponent=self.e)

            assert key._d == critter.expected_d


class TestSage:
    """Test data from SageMath Tutorial

    https://doc.sagemath.org/html/en/thematic_tutorials/numtheory_rsa.html

    The tutorial correctly points out that they way the primes
    were generated is inappropriate for real work.

    The tutorial uses phi directly instead of lcm(p-1, q-1).
    """

    # Don't use Mersenne primes in real life
    p = (2**31) - 1
    q = (2**61) - 1
    e = 1850567623300615966303954877
    m = 72697676798779827668  # message

    n = 4951760154835678088235319297
    phi = 4951760152529835076874141700
    d = 4460824882019967172592779313
    c = 630913632577520058415521090

    λ = lcm(p - 1, q - 1)

    def test_encrypt(self) -> None:
        priv_key = rsa.PrivateKey(self.p, self.q, self.e)
        pub_key = priv_key.pub_key

        assert pub_key.encrypt(self.m) == self.c

    def test_decrypt(self) -> None:
        priv_key = rsa.PrivateKey(self.p, self.q, self.e)

        assert priv_key.decrypt(self.c) == self.m

    def test_N(self) -> None:
        priv_key = rsa.PrivateKey(self.p, self.q, self.e)
        pub_key = priv_key.pub_key

        assert self.n == pub_key.N

    def test_d(self) -> None:
        priv_key = rsa.PrivateKey(self.p, self.q, self.e)

        # We (almost certainly) get a smaller d where the lcm check matters
        if self.phi == self.λ:
            assert priv_key._d == self.d


class TestMG1977:
    # encoder/decoder is in utils.Rsa129
    def test_magic(self) -> None:
        """Test the RSA-129 Challenge from Martin Gardner's 1977 article"""

        Challenge = namedtuple(
            "Challenge", ["modulus", "pub_exponent", "ctext"]
        )
        Solution = namedtuple("Solution", ["p", "q", "plaintext"])

        # From Martin Gardner's 1977
        challenge = Challenge(
            modulus=114381625757888867669235779976146612010218296721242362562561842935706935245733897830597123563958705058989075147599290026879543541,
            pub_exponent=9007,
            ctext=96869613754622061477140922254355882905759991124574319874695120930816298225145708356931476622883989628013391990551829945157815154,
        )

        # From Atkins et al 1995
        solution = Solution(
            p=3490529510847650949147849619903898133417764638493387843990820577,
            q=32769132993266709549961988190834461413177642967992942539798288533,
            plaintext="THE MAGIC WORDS ARE SQUEAMISH OSSIFRAGE",
        )

        pub_key = rsa.PublicKey(challenge.modulus, challenge.pub_exponent)

        # First test encryption
        plain_num: int = Rsa129.encode(solution.plaintext)
        ctext = pub_key.encrypt(plain_num)
        assert ctext == challenge.ctext

        # Now test decryption
        priv_key = rsa.PrivateKey(
            solution.p, solution.q, challenge.pub_exponent
        )
        assert priv_key == pub_key

        decrypted_num: int = priv_key.decrypt(ctext)
        decrypted: str = Rsa129.decode(decrypted_num)

        assert decrypted == solution.plaintext


class TestEq:
    # We will reuse magic values for this test, but anything should do
    p = 3490529510847650949147849619903898133417764638493387843990820577
    q = 32769132993266709549961988190834461413177642967992942539798288533
    e = 9007

    def test_pq_order(self) -> None:
        priv_pq = rsa.PrivateKey(self.p, self.q)
        priv_qp = rsa.PrivateKey(self.q, self.p)

        assert priv_pq == priv_qp

    def test_pub(self) -> None:
        priv_pq = rsa.PrivateKey(self.p, self.q, pub_exponent=self.e)
        priv_qp = rsa.PrivateKey(self.q, self.p, pub_exponent=self.e)

        assert priv_pq == priv_qp

    def test_pub_phi(self) -> None:
        """Key computed with 𝜑 is equivalent to key computed with λ."""

        key_lambda = rsa.PrivateKey(self.p, self.q, self.e)

        # can only construct key_phi by changing what would be private fields
        key_phi = rsa.PrivateKey(self.p, self.q, self.e)

        phi = (self.p - 1) * (self.q - 1)
        d_phi = modinv(key_phi.e, phi)

        key_phi._d = d_phi

        assert key_phi == key_lambda


class TestMisc:
    p = 3490529510847650949147849619903898133417764638493387843990820577
    q = 32769132993266709549961988190834461413177642967992942539798288533

    def test_default_e(self) -> None:
        default_e = rsa.default_e()
        assert default_e == 65537

        priv_key = rsa.PrivateKey(self.p, self.q)

        assert priv_key.e == default_e


class TestOaep:
    # from 1024 bit key at
    # https://github.com/pyca/cryptography/blob/main/vectors/cryptography_vectors/asymmetric/RSA/pkcs-1v2-1d2-vec/oaep-vect.txt

    prime1 = bytes.fromhex("""
                d3 27 37 e7 26 7f fe 13 41 b2 d5 c0 d1 50 a8 1b
                58 6f b3 13 2b ed 2f 8d 52 62 86 4a 9c b9 f3 0a
                f3 8b e4 48 59 8d 41 3a 17 2e fb 80 2c 21 ac f1
                c1 1c 52 0c 2f 26 a4 71 dc ad 21 2e ac 7c a3 9d
            """)

    prime2 = bytes.fromhex("""
            cc 88 53 d1 d5 4d a6 30 fa c0 04 f4 71 f2 81 c7
            b8 98 2d 82 24 a4 90 ed be b3 3d 3e 3d 5c c9 3c
            47 65 70 3d 1d d7 91 64 2f 1f 11 6a 0d d8 52 be
            24 19 b2 af 72 bf e9 a0 30 e8 60 b0 28 8b 5d 77
        """)

    exponent = 65537

    # test vectors from https://github.com/bdauvergne/python-pkcs1/
    pkcs_vector: dict[str, bytes] = {
        "prime1": bytes.fromhex(
            "ee cf ae 81 b1 b9 b3 c9 08 81 0b 10 a1 b5 60 01"
            "99 eb 9f 44 ae f4 fd a4 93 b8 1a 9e 3d 84 f6 32"
            "12 4e f0 23 6e 5d 1e 3b 7e 28 fa e7 aa 04 0a 2d"
            "5b 25 21 76 45 9d 1f 39 75 41 ba 2a 58 fb 65 99"
        ),
        "prime2": bytes.fromhex(
            "c9 7f b1 f0 27 f4 53 f6 34 12 33 ea aa d1 d9 35"
            "3f 6c 42 d0 88 66 b1 d0 5a 0f 20 35 02 8b 9d 86"
            "98 40 b4 16 66 b4 2e 92 ea 0d a3 b4 32 04 b5 cf"
            "ce 33 52 52 4d 04 16 a5 a4 41 e7 00 af 46 15 03"
        ),
        "exponent_bytes": bytes([0x11]),
        "message": bytes.fromhex(
            "d4 36 e9 95 69 fd 32 a7c8 a0 5b bc 90 d3 2c 49"
        ),
        "seed": bytes.fromhex(
            "aa fd 12 f6 59 ca e6 34 89 b479 e5 07 6d de c2 f0 6c b5 8f"
        ),
        "seed_mask": bytes.fromhex(
            "41 87 0b 5a b0 29 e6 57 d9 57 50 b5 4c 28 3c 08 72 5d be a9"
        ),
        "ciphertext": bytes.fromhex(
            "12 53 e0 4d c0 a5 39 7b b4 4a 7a b8 7e 9b f2 a0 39 a3 3d 1e"
            "99 6f c8 2a 94 cc d3 00 74 c9 5d f7 63 72 20 17 06 9e 52 68"
            "da 5d 1c 0b 4f 87 2c f6 53 c1 1d f8 23 14 a6 79 68 df ea e2"
            "8d ef 04 bb 6d 84 b1 c3 1d 65 4a 19 70 e5 78 3b d6 eb 96 a0"
            "24 c2 ca 2f 4a 90 fe 9f 2e f5 c9 c1 40 e5 bb 48 da 95 36 ad"
            "87 00 c8 4f c9 13 0a de a7 4e 55 8d 51 a7 4d df 85 d8 b5 0d"
            "e9 68 38 d6 06 3e 09 55"
        ),
        "db": bytes.fromhex(
            "da 39 a3 ee 5e 6b 4b 0d 32 55 bf ef 95 60 18 90 af d8 07 09"
            "00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00"
            "00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00"
            "00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00"
            "00 00 00 00 00 00 00 00 00 00 01 d4 36 e9 95 69 fd 32 a7 c8"
            "a0 5b bc 90 d3 2c 49"
        ),
        "db_mask": bytes.fromhex(
            "06 e1 de b2 36 9a a5 a5 c7 07 d8 2c 8e 4e 93 24 8a c7 83 de"
            "e0 b2 c0 46 26 f5 af f9 3e dc fb 25 c9 c2 b3 ff 8a e1 0e 83"
            "9a 2d db 4c dc fe 4f f4 77 28 b4 a1 b7 c1 36 2b aa d2 9a b4"
            "8d 28 69 d5 02 41 21 43 58 11 59 1b e3 92 f9 82 fb 3e 87 d0"
            "95 ae b4 04 48 db 97 2f 3a c1 4e af f4 9c 8c 3b 7c fc 95 1a"
            "51 ec d1 dd e6 12 64"
        ),
        "masked_db": bytes.fromhex(
            "dc d8 7d 5c 68 f1 ee a8 f5 52 67 c3 1b 2e 8b b4 25 1f 84 d7"
            "e0 b2 c0 46 26 f5 af f9 3e dc fb 25 c9 c2 b3 ff 8a e1 0e 83"
            "9a 2d db 4c dc fe 4f f4 77 28 b4 a1 b7 c1 36 2b aa d2 9a b4"
            "8d 28 69 d5 02 41 21 43 58 11 59 1b e3 92 f9 82 fb 3e 87 d0"
            "95 ae b4 04 48 db 97 2f 3a c1 4f 7b c2 75 19 52 81 ce 32 d2"
            "f1 b7 6d 4d 35 3e 2d"
        ),
    }

    key1 = rsa.PrivateKey(
        int.from_bytes(prime1, byteorder="big"),
        int.from_bytes(prime2, byteorder="big"),
        pub_exponent=exponent,
    )

    def test_enc(self) -> None:
        class Vector:
            def __init__(
                self, name: str, message: str, seed: str, encryption: str
            ) -> None:
                self.name = name
                self.message = bytes.fromhex(message)
                self.seed = bytes.fromhex(seed)
                self.encryption = bytes.fromhex(encryption)

        vectors: list[Vector] = [
            Vector(
                name="Example 1.1",
                message="""
                    66 28 19 4e 12 07 3d b0 3b a9 4c da 9e f9 53 23
                    97 d5 0d ba 79 b9 87 00 4a fe fe 34
                """,
                seed="""
                    18 b7 76 ea 21 06 9d 69 77 6a 33 e9 6b ad 48 e1
                    dd a0 a5 ef 
                """,
                encryption="""
                    35 4f e6 7b 4a 12 6d 5d 35 fe 36 c7 77 79 1a 3f
                    7b a1 3d ef 48 4e 2d 39 08 af f7 22 fa d4 68 fb
                    21 69 6d e9 5d 0b e9 11 c2 d3 17 4f 8a fc c2 01
                    03 5f 7b 6d 8e 69 40 2d e5 45 16 18 c2 1a 53 5f
                    a9 d7 bf c5 b8 dd 9f c2 43 f8 cf 92 7d b3 13 22
                    d6 e8 81 ea a9 1a 99 61 70 e6 57 a0 5a 26 64 26
                    d9 8c 88 00 3f 84 77 c1 22 70 94 a0 d9 fa 1e 8c
                    40 24 30 9c e1 ec cc b5 21 00 35 d4 7a c7 2e 8a 
                """,
            )
        ]

        for v in vectors:
            ctext = self.key1.pub_key.oaep_encrypt(
                v.message, hash_id="sha1", mgf_id="mgf1SHA1", _seed=v.seed
            )
            assert ctext == v.encryption

    def test_enc_dec(self) -> None:
        p = int.from_bytes(self.prime1, "big")
        q = int.from_bytes(self.prime2, "big")
        key = rsa.PrivateKey(p, q, pub_exponent=self.exponent)

        plaintext = b"THE MAGIC WORDS ARE SQUEAMISH OSSIFRAGE"

        # As a sanity check, do primitive RSA with this key
        int_ptext = int.from_bytes(plaintext, "big")

        primitive_ctext = key.pub_key.encrypt(int_ptext)
        primitive_decrypted = key.decrypt(primitive_ctext)
        assert primitive_decrypted == int_ptext

        ctext = key.pub_key.oaep_encrypt(plaintext)
        decrypted = key.oaep_decrypt(ctext)

        assert plaintext == decrypted

    def test_mgf1(self) -> None:
        seed = self.pkcs_vector["seed"]
        db = self.pkcs_vector["db"]
        db_mask = rsa.Oaep.mgf1(seed, len(db), "sha1")
        assert db_mask == self.pkcs_vector["db_mask"]

    def test_enc_pkcs(self) -> None:
        """Test with vectors from Python PKCS"""

        p1 = int.from_bytes(self.pkcs_vector["prime1"], "big")
        p2 = int.from_bytes(self.pkcs_vector["prime2"], "big")
        e = int.from_bytes(self.pkcs_vector["exponent_bytes"], "big")

        key = rsa.PrivateKey(p1, p2, pub_exponent=e)
        pub_key = key.pub_key

        m = self.pkcs_vector["message"]
        ctext = pub_key.oaep_encrypt(
            m,
            hash_id="sha1",
            mgf_id="mgf1SHA1",
            _seed=self.pkcs_vector["seed"],
        )
        decrypted = key.oaep_decrypt(ctext, hash_id="sha1", mgf_id="mgf1SHA1")

        assert m == decrypted
        assert ctext == self.pkcs_vector["ciphertext"]



# SPDX-FileCopyrightText: 2024-present Jeffrey Goldberg <jeffrey@goldmark.org>
#
# SPDX-License-Identifier: MIT

import math
from collections import UserList
from collections.abc import Iterator, Iterable
from typing import Any, NewType, Optional, Self, TypeGuard

import primefac

try:
    from bitarray import bitarray
    from bitarray.util import count_n, ba2int
except ImportError:

    def bitarray(*args, **kwargs) -> Any:  # type: ignore
        raise NotImplementedError("bitarray is not installed")

    def count_n(*args, **kwargs) -> int:  # type: ignore
        raise NotImplementedError("bitarray is not installed")

    def ba2int(*args, **kwargs) -> int:  # type: ignore
        raise NotImplementedError("bitarray is not installed")


from . import types


Modulus = NewType("Modulus", int)
"""type Modulus is an int greater than 1"""


def is_modulus(n: Any) -> TypeGuard[Modulus]:
    if not isinstance(n, int):
        return False
    if n < 2:
        return False
    if not isprime(n):
        return False
    return True


def isprime(n: int) -> bool:
    """False if composite; True if very probably prime."""
    return primefac.isprime(n)


def isqrt(n: int) -> int:
    """returns the greatest r such that r * r =< n"""
    if n < 0:
        raise ValueError("n cannot be negative")
    return math.isqrt(n)


def modinv(a: int, m: int) -> int:
    """
    Returns b such that :math:`ab \\equiv 1 \\pmod m`.

    :raises ValueError: if a is not coprime with m
    """

    # python 3.8 allows -1 as power.
    return pow(a, -1, m)


class FactorList(UserList[tuple[int, int]]):
    """
    A FactorList is an list of (prime, exponent) tuples.

    It represents the prime factorization of a number.
    """

    def __init__(
        self,
        prime_factors: list[tuple[int, int]] = [],
        check_primes: bool = False,
    ) -> None:
        """
        prime_factors should be a list of (prime, exponent) tuples.

        Either you ensure that the primes really are prime or use
        ``check_primes = True``.
        """
        super().__init__(prime_factors)

        # Normalization will do some sanity checking as well
        self.normalize(check_primes=check_primes)

        # property-like things that are computed when first needed
        self._n: Optional[int] = None
        self._totient: Optional[int] = None
        self._radical: Optional[FactorList] = None
        self._radical_value: Optional[int] = None
        self._factors_are_prime: Optional[bool] = None

    def __repr__(self) -> str:
        s: list[str] = []
        for p, e in self.data:
            term = f"{p}" if e == 1 else f"{p}^{e}"
            s.append(term)
        return " * ".join(s)

    def __eq__(self, other: object) -> bool:
        # Implemented for
        #  - list
        #  - int
        #  - UserDict
        if isinstance(other, list):
            try:
                other_f = FactorList(other)
            except (ValueError, TypeError):
                return False
            return self.data == other_f.data

        # Fundamental theorem of arithmetic
        if isinstance(other, int):
            return self.n == other

        if not isinstance(other, UserList):
            return NotImplemented

        return self.data == other.data

    def __add__(self, other: Iterable[tuple[int, int]]) -> "FactorList":
        added = super().__add__(other)
        added = FactorList(added.data)  # init will normalize
        return added

    def normalize(self, check_primes: bool = False) -> Self:
        """
        Deduplicates and sorts in prime order, removing exponent == 0 cases.

        :raises TypeError: if prime and exponents are not ints

        :raises ValueError: if p < 2 or e < 0

        This only checks that primes are prime if ``check_primes`` is True.

        """

        # this calls for some clever list comprehensions.
        # But I am not feeling that clever at the moment

        # I will construct a dict from the data and then
        # reconstruct the data from the dict

        d = {p: 0 for (p, _) in self.data}
        for p, e in self.data:
            if not isinstance(p, int) or not isinstance(e, int):
                raise TypeError("Primes and exponents must be integers")
            if p < 2:
                raise ValueError(f"{p} should be greater than 1")
            if e == 0:
                continue
            if e < 0:
                raise ValueError(f"exponent ({e}) should not be negative")
            if check_primes:
                if not isprime(p):
                    raise ValueError(f"{p} is composite")
            d[p] += e

        self.data = [(p, d[p]) for p in sorted(d.keys())]

        return self

    @property
    def factors_are_prime(self) -> bool:
        """True iff all the alleged primes are prime."""
        if self._factors_are_prime is not None:
            return self._factors_are_prime
        self._factors_are_prime = all([isprime(p) for p, _ in self.data])
        return self._factors_are_prime

    @property
    def n(self) -> int:
        """The integer that this is a factorization of"""
        if self._n is None:
            self._n = int(math.prod([p**e for p, e in self.data]))
        return self._n

    @property
    def phi(self) -> int:
        """
        Returns Euler's Totient (phi)

        :math:`\\phi(n)` is the number of numbers
        less than n which are coprime with n.

        This assumes that the factorization (self) is a prime factorization.
        """

        if self._totient is None:
            self._totient = int(
                math.prod([p ** (e - 1) * (p - 1) for p, e in self.data])
            )

        return self._totient

    def coprimes(self) -> Iterator[int]:
        """Iterator of coprimes."""
        for a in range(1, self.n):
            if not any([a % p == 0 for p, _ in self.data]):
                yield a

    def unit(self) -> int:
        """Unit is always 1 for positive integer factorization."""
        return 1

    def is_integral(self) -> bool:
        """Always true for integer factorization."""
        return True

    def value(self) -> int:
        """Same as ``n()``."""
        return self.n

    def radical(self) -> "FactorList":
        """All exponents on factors set to 1"""
        if self._radical is None:
            self._radical = FactorList([(p, 1) for p, _ in self.data])
        return self._radical

    def radical_value(self) -> int:
        """Product of factors each used only once."""
        if self._radical_value is None:
            self._radical_value = math.prod([p for p, _ in self.data])
        return self._radical_value

    def pow(self, n: int) -> "FactorList":
        """Return (self)^n, where *n* is positive int."""
        if not types.is_positive_int(n):
            raise TypeError("n must be a positive integer")

        return FactorList([(p, n * e) for p, e in self.data])


def factor(n: int, ith: int = 0) -> FactorList:
    """
    Returns list (prime, exponent) factors of n.
    Starts trial div at ith prime.

    This wraps ``primefac.primefac()``, but creates our FactorList
    """

    primes = primefac.primefac(n)

    return FactorList([(p, 1) for p in primes])


def gcd(*integers: int) -> int:
    """Returns greatest common denominator of arguments."""
    return math.gcd(*integers)


def egcd(a: int, b: int) -> tuple[int, int, int]:
    """returns (g, x, y) such that :math:`ax + by = \\gcd(a, b) = g`."""
    x0, x1, y0, y1 = 0, 1, 1, 0
    while a != 0:
        (q, a), b = divmod(b, a), a
        y0, y1 = y1, y0 - q * y1
        x0, x1 = x1, x0 - q * x1
    return b, x0, y0


def is_square(n: int) -> bool:
    """True iff n is a perfect square."""

    return primefac.ispower(n, 2) is not None


def mod_sqrt(a: int, m: int) -> list[int]:
    """Modular square root.

    For prime m, this generally returns a list with either two members,
    :math:`[r, m - r]` such that :math:`r^2 = {(m - r)}^2 = a \\pmod m`
    if such an a is a quadratic residue the empty list if there is no such r.

    However, for compatibility with SageMath this can return a list with just
    one element in some special cases

    - returns ``[0]`` if :math:`m = 3` or :math:`a = 0`.
    - returns ``[a % m]`` if :math:`m = 2`.

    Otherwise it returns a list with the two quadratic residues if they exist
    or and empty list otherwise.

    **Warning**: The behavior is undefined if m is not prime.
    """

    match m:
        case 2:
            return [a % m]
        case 3:
            return [0]
        case _ if m < 2:
            raise ValueError("modulus must be prime")

    if a == 1:
        return [1, m - 1]

    a = a % m
    if a == 0:
        return [0]

    # check that a is a quadratic residue, return []] if not
    if pow(a, (m - 1) // 2, m) != 1:
        return []

    v = primefac.sqrtmod_prime(a, m)
    return [v, (m - v) % m]


def lcm(*integers: int) -> int:
    """Least common multiple"""

    # requires python 3.9, but I'm already requiring 3.11
    return math.lcm(*integers)



import sys

import pytest
from collections.abc import Iterator
from toy_crypto import utils
from toy_crypto.types import Byte


class TestUtils:
    def test_digit_count(self) -> None:
        i_e61 = pow(10, 61)  # because '1e61` does floating point operations
        vectors = [
            (999, 3),
            (1000, 4),
            (1001, 4),
            (i_e61, 62),
            (i_e61 - 1, 61),
            (i_e61 + 1, 62),
            (-999, 3),
            (0, 1),
            (-0, 1),
        ]
        for n, expected in vectors:
            d = utils.digit_count(n)
            assert d == expected


class TestXor:
    def test_xor(self) -> None:
        vectors = [
            (b"dusk", b"dawn", bytes.fromhex("00 14 04 05")),
            (
                b"Attack at dawn!",
                bytes(10) + bytes.fromhex("00 14 04 05 00"),
                b"Attack at dusk!",
            ),
            (
                bytes(15),
                bytes.fromhex("00 01 02"),
                bytes.fromhex("00 01 02") * 5,
            ),
        ]

        for x, y, pad in vectors:
            r = utils.xor(x, y)
            assert r == pad

    def test_iter(self) -> None:
        pad = bytes.fromhex("00 36 5C")
        p_modulus = len(pad)
        single = list(range(0, 256))
        s_modulus = len(single)
        message: Iterator[Byte] = iter(single * 10)

        iter_xor = utils.Xor(message, pad)

        m_idx = 0
        p_idx = 0
        for b in iter_xor:
            expected = (m_idx % s_modulus) ^ pad[p_idx]
            assert b == expected
            m_idx += 1
            p_idx = (p_idx + 1) % p_modulus

        assert m_idx == s_modulus * 10


class TestRsa29Encoding:
    vectors: list[tuple[str, int]] = [
        ("ITS ALL GREEK TO ME", 9201900011212000718050511002015001305),
        (
            "THE MAGIC WORDS ARE SQUEAMISH OSSIFRAGE",
            200805001301070903002315180419000118050019172105011309190800151919090618010705,
        ),
    ]

    def test_encode(self) -> None:
        for s, n in self.vectors:
            encoded = utils.Rsa129.encode(s)
            assert encoded == n

    def test_decode(self) -> None:
        for s, n in self.vectors:
            decoded = utils.Rsa129.decode(n)
            assert decoded == s


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



import os
import sys
import time

PROJECT_PATH = os.getcwd()
SOURCE_PATH = os.path.join(PROJECT_PATH, "src")
sys.path.append(SOURCE_PATH)

from toy_crypto.sieve import BaSieve  # noqa: E402

setup = """
from toy_crypto.sieve import Sieve
"""

FIRST_SIZE = 10_000_000
FINAL_SIZE = 100_000_000
TRIALS = 10

PRIME = 71
LEN = 100_000


def f1() -> None:
    BaSieve._reset()
    s1 = BaSieve.from_size(FIRST_SIZE)
    assert s1.count == 664579
    s2 = BaSieve.from_size(FINAL_SIZE)
    assert s2.count == 5761455


def f2() -> None:
    BaSieve._reset()
    s1 = BaSieve.from_size(FIRST_SIZE)
    assert s1.count == 664579
    BaSieve._reset()
    s2 = BaSieve.from_size(FINAL_SIZE)
    assert s2.count == 5761455


def f3() -> None:
    BaSieve._reset()
    s1 = BaSieve.from_size(FIRST_SIZE)
    s2 = BaSieve.from_size(FIRST_SIZE)
    assert s1.count == s2.count


def f4() -> None:
    BaSieve._reset()
    s1 = BaSieve.from_size(FIRST_SIZE)
    BaSieve._reset()
    s2 = BaSieve.from_size(FIRST_SIZE)
    assert s1.count == s2.count


def main() -> None:
    trials = 5

    t_f2_start = time.time()
    for _ in range(trials):
        f2()
    delta_f2 = time.time() - t_f2_start

    t_f1_start = time.time()
    for _ in range(trials):
        f1()
    delta_f1 = time.time() - t_f1_start

    print(
        f"For {FIRST_SIZE}, {FINAL_SIZE}\n"
        f"\t{delta_f1}/{delta_f2} = {delta_f1 / delta_f2:.2f}"
    )

    t_f3_start = time.time()
    for _ in range(trials):
        f3()
    delta_f3 = time.time() - t_f3_start

    t_f4_start = time.time()
    for _ in range(trials):
        f4()
    delta_f4 = time.time() - t_f4_start

    print(
        f"For {FIRST_SIZE}, {FIRST_SIZE}\n"
        f"\t{delta_f3}/{delta_f4} = {delta_f3 / delta_f4:.2f}"
    )


if __name__ == "__main__":
    main()



import base64
import sys
from pathlib import Path

import pytest
from toy_crypto import vigenere

probabistic = pytest.mark.skip(reason="Probabilistic")

Letter = vigenere.Letter


class TestCryptionDefault:
    """Encryption/decryption tests with default alphabet."""

    default_abc = None

    class TVector:
        def __init__(self, key: str, ptext: str, ctext: str) -> None:
            self.key = key
            self.ptext = ptext
            self.ctext = ctext

    vectors = [
        TVector(
            "AAAA",
            "ONCE UPON A MIDNIGHT DREARY",
            "ONCE UPON A MIDNIGHT DREARY",
        ),
        TVector(
            "C",
            "ONCE UPON A MIDNIGHT DREARY",
            "QPEG WRQP C OKFPKIJV FTGCTA",
        ),
        TVector(
            "RAVEN",
            "ONCE UPON A MIDNIGHT DREARY",
            "FNXI HGOI E ZZDIMTYT YVRRRT",
        ),
        TVector(
            "RAVEN",
            "AAAA AAAA A AAAAAAAA AAAAAA",
            "RAVE NRAV E NRAVENRA VENRAV",
        ),
    ]

    def test_encrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.default_abc)
            ctext = cipher.encrypt(tv.ptext)

            assert ctext == tv.ctext

    def test_decrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.default_abc)
            ptext = cipher.decrypt(tv.ctext)

            assert ptext == tv.ptext


class TestCryptionCapsOnly:
    """Encryption/decryption tests with CAPS_ONLY alphabet."""

    default_abc = vigenere.Alphabet.CAPS_ONLY

    class TVector:
        def __init__(self, key: str, ptext: str, ctext: str) -> None:
            self.key = key
            self.ptext = ptext
            self.ctext = ctext

    vectors = [
        TVector(
            "AAAA",
            "ONCE UPON A MIDNIGHT DREARY",
            "ONCE UPON A MIDNIGHT DREARY",
        ),
        TVector(
            "C",
            "ONCE UPON A MIDNIGHT DREARY",
            "QPEG WRQP C OKFPKIJV FTGCTA",
        ),
        TVector(
            "RAVEN",
            "ONCE UPON A MIDNIGHT DREARY",
            "FNXI HGOI E ZZDIMTYT YVRRRT",
        ),
        TVector(
            "RAVEN",
            "AAAA AAAA A AAAAAAAA AAAAAA",
            "RAVE NRAV E NRAVENRA VENRAV",
        ),
    ]

    def test_encrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.default_abc)
            ctext = cipher.encrypt(tv.ptext)

            assert ctext == tv.ctext

    def test_decrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.default_abc)
            ptext = cipher.decrypt(tv.ctext)

            assert ptext == tv.ptext


class TestCryptionPrintable:
    """Encryption/decryption tests with printable alphabet."""

    abc = vigenere.Alphabet(prebaked="printable")

    class TVector:
        def __init__(self, key: str, ptext: str, ctext: str) -> None:
            self.key = key
            self.ptext = ptext
            self.ctext = ctext

    vectors = [
        TVector(
            abc.alphabet[0],
            "ONCE UPON A MIDNIGHT DREARY",
            "ONCE UPON A MIDNIGHT DREARY",
        ),
        TVector(
            "C",
            "ONCE UPON A MIDNIGHT DREARY",
            r"""VBryLMgVBL;LZ1FB1[UILF?y;?b""",
        ),
        TVector(
            "RAVEN",
            "ONCE UPON A MIDNIGHT DREARY",
            r"""gVP=:,+YA:xX$,+2&;'3=.G=VU7""",
        ),
    ]

    def test_encrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.abc)
            ctext = cipher.encrypt(tv.ptext)

            assert ctext == tv.ctext

    def test_decrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.abc)
            ptext = cipher.decrypt(tv.ctext)

            assert ptext == tv.ptext


class TestCryptionCustomAbc:
    """Encryption/decryption tests with custom alphabet."""

    abc = vigenere.Alphabet("0123456789ABCDEF")

    class TVector:
        def __init__(self, key: str, ptext: str, ctext: str) -> None:
            self.key = key
            self.ptext = ptext
            self.ctext = ctext

    vectors = [
        TVector(
            abc[0],
            "DEAD BEEF BAD 5EED",
            "DEAD BEEF BAD 5EED",
        ),
        TVector(
            "2",
            "DEAD BEEF BAD 5EED",
            "F0CF D001 DCF 700F",
        ),
        TVector(
            "BADA55",
            "DEAD BEEF BAD 5EED",
            "8877 0399 842 A98A",
        ),
    ]

    def test_encrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.abc)
            ctext = cipher.encrypt(tv.ptext)

            assert ctext == tv.ctext

    def test_decrypt(self) -> None:
        for tv in self.vectors:
            cipher = vigenere.Cipher(tv.key, self.abc)
            ptext = cipher.decrypt(tv.ctext)

            assert ptext == tv.ptext


class TestCrack:
    cryptopals_filename = "tests/s1c6.txt"
    _encoded = Path(cryptopals_filename).read_text(encoding="ASCII")
    cp_ciphertext: bytes = base64.b64decode(_encoded)
    cp_key = b"TerMinator X: Bring the noiSe"

    @probabistic
    def test_find_keysize(self) -> None:
        # We set number of trial pairs high to reduce chance random failure
        trial_pairs = 8

        scores = vigenere.probable_keysize(
            self.cp_ciphertext,
            min_size=7,
            max_size=35,
            trial_pairs=trial_pairs,
        )

        expected = len(self.cp_key)
        best_length = scores.best
        assert best_length == expected


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



import math

from . import types


MAX_QBIRTHDAY_P = 1.0 - (10**-8)
"""Maximum probability that Q can handle."""

EXACT_THRESHOLD = 1000
"""With auto mode, the threshold for using exact or approximate modes."""


def _pbirthday_exact(
    n: types.PositiveInt, classes: types.PositiveInt, coincident: int
) -> types.Prob:
    # use notation  from Diconis and Mosteller 1969
    c = classes  # classes
    k = coincident

    if k < 2:
        return types.Prob(1.0)
    if k > 2:
        return _pbirthday_approx(n, c, coincident=k)

    if n >= c:
        return types.Prob(1.0)

    v_dn = math.perm(c, n)
    v_t = pow(c, n)

    p = 1.0 - float(v_dn / v_t)
    if not types.is_prob(p):
        raise Exception("this should not happen")
    return p


def _pbirthday_approx(
    n: types.PositiveInt, classes: types.PositiveInt, coincident: int
) -> types.Prob:
    # DM1969 notation
    c = classes
    k = coincident

    if n >= c * (k - 1):
        return types.Prob(1.0)

    if k < 2:
        return types.Prob(1.0)

    # p = 1.0 - math.exp(-(n * n) / (2 * d))

    # lifted from R src/library/stats/R/birthday.R
    LHS = n * math.exp(-n / (c * k)) / (1 - n / (c * (k + 1))) ** (1 / k)
    lxx = k * math.log(LHS) - (k - 1) * math.log(c) - math.lgamma(k + 1)
    p = -math.expm1(-math.exp(lxx))
    if not types.is_prob(p):
        raise Exception("this should not happen")
    return p


def P(
    n: int, classes: int = 365, coincident: int = 2, mode: str = "auto"
) -> types.Prob:
    """probability of at least 1 collision among n individuals for c classes".

    The "exact" method still involves floating point approximations
    and may be very slow for large n.

    :raises ValueError: if any of :data:`n`, :data:`classes`,
        or :data:`coincident` are less than 1.
    """
    c = classes
    k = coincident

    if not types.is_positive_int(n):
        raise ValueError("n must be a positive integer")
    if not types.is_positive_int(c):
        raise ValueError("classes must be a positive integer")
    if not types.is_positive_int(k):
        raise ValueError("coincident must be a positive integer")

    if k == 1:
        return types.Prob(1.0)

    if mode == "auto":
        mode = "exact" if c < EXACT_THRESHOLD else "approximate"
    match mode:
        case "exact":
            return _pbirthday_exact(n, c, coincident=k)
        case "approximate":
            return _pbirthday_approx(n, c, coincident=k)
        case _:
            raise ValueError('mode must be "auto", "exact", or  "approximate"')


def Q(prob: float = 0.5, classes: int = 365, coincident: int = 2) -> int:
    """Returns minimum number n to get a probability of p for c classes.

    :raises ValueError: if :data:`prob` is less than 0 or greater than 1.
    :raises ValueError: if :data:`classes` is less than 1.
    :raises ValueError: if :data:`coincident` is less than 1.
    """

    if not types.is_prob(prob):
        raise ValueError(f"{prob} is not a valid probability")

    if classes < 1:
        raise ValueError("classes must be positive")
    if coincident < 1:
        raise ValueError("coincident must be positive")

    # Use DM69 notation so I can better connect code to published method.
    p = prob
    c = classes
    k = coincident

    if p > MAX_QBIRTHDAY_P:
        return c * (k - 1) + 1

    # Lifted from R src/library/stats/R/birthday.R
    if p == types.Prob(0):
        return 1

    # First approximation
    # broken down so that I can better understand this.
    term1 = (k - 1) * math.log(c)  # log(c^{k-1})
    term2 = math.lgamma(k + 1)  # log k!
    term3 = math.log(-math.log1p(-p))  # ?
    log_n = (term1 + term2 + term3) / k  # adding log x_i is log prod x_i
    n = math.exp(log_n)
    n = math.ceil(n)

    if P(n, c, coincident=k) < p:
        n += 1
        while P(n, c, coincident=k) < p:
            n += 1
    elif P(n - 1, c, coincident=k) >= p:
        n -= 1
        while P(n - 1, c, coincident=k) >= p:
            n -= 1

    return n



import sys

import pytest
from toy_crypto import rand


class TestRandRange:
    def test_range_even(self) -> None:
        expected_values = range(0, 10, 2)
        counts: dict[int, int] = {r: 0 for r in expected_values}
        for _ in range(1000):
            r = rand.randrange(0, 10, 2)
            assert r in expected_values

            counts[r] += 1

        # could run a more sophisticated statistical test, but
        # let's start with this
        assert all([c > 0 for c in counts.values()])

    def test_single_arg(self) -> None:
        top = 20
        trials = 200 * top
        counts: dict[int, int] = {r: 0 for r in range(top)}

        for _ in range(trials):
            r = rand.randrange(top)
            assert r >= 0
            assert r < top

            counts[r] += 1

        assert all([c > 0 for c in counts.values()])


def test_random() -> None:
    trials = 1024
    for _ in range(trials):
        x = rand.random()
        assert x >= 0.0
        assert x < 1.0


# Slightly different invocations of mypy have very different opinions
@pytest.mark.statistical  # type: ignore[misc,unused-ignore]
def test_shuffle() -> None:
    trials = 2048

    input = ["A", "B", "C"]
    permutations = ["ABC", "ACB", "BAC", "BCA", "CAB", "CBA"]
    counts = {p: 0 for p in permutations}

    for _ in range(trials):
        # Note that shuffle shuffles in place
        c = input.copy()
        rand.shuffle(c)
        s = "".join(c)
        assert s in permutations
        counts[s] += 1

    mle = trials // len(permutations)
    # I should calculate the odds of test failure for good shuffle, but not now
    expected_floor = int(mle / 2)
    expected_ceiling = trials - expected_floor

    assert min(counts.values()) > expected_floor
    assert max(counts.values()) < expected_ceiling


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



# mypy: disable-error-code="type-abstract"
import sys

import pytest
from toy_crypto import sieve


class Fixed:
    """Perhaps better done with fixtures"""

    expected30_int = 545925292
    """stringy bitarray for primes below 30"""

    ints: list[tuple[int, int]] = [
        (30, int("100000100010100010100010101100", 2)),  # 545925292
        (100, 159085582874019712269820766380),
    ]
    primes100: list[int] = [
        2, 3, 5, 7, 11, 13, 17, 19,
        23, 29, 31, 37, 41, 43, 47,
        53, 59, 61, 67, 71, 73,
        79, 83, 89, 97,
    ]  # fmt: skip
    """primes below 100"""

    primes30: list[int] = [
        2, 3, 5, 7, 11, 13, 17, 19,
        23, 29,
    ]  # fmt: skip
    """primes below 30"""

    sc: sieve.Sievish

    @classmethod
    def t_30(cls, sc: type[sieve.Sievish]) -> None:
        sc._reset()
        s30 = sc.from_size(30)
        s30_count = 10

        assert int(s30) == cls.expected30_int
        assert s30_count == s30.count

    @classmethod
    def t_count(cls, sc: type[sieve.Sievish]) -> None:
        s100 = sc.from_size(100)
        result = s100.count
        assert result == len(cls.primes100)

    @classmethod
    def t_primes(cls, sc: type[sieve.Sievish]) -> None:
        sc._reset()
        s30 = sc.from_size(30)
        expected = cls.primes30

        primes = list(s30.primes())
        assert primes == expected

    @classmethod
    def t_2int(cls, sc: type[sieve.Sievish]) -> None:
        for size, expected in cls.ints:
            s = sc.from_size(size)
            i = int(s)
            assert i == expected

    @classmethod
    def t_from_int(cls, sc: type[sieve.Sievish]) -> None:
        for _, t_int in cls.ints:
            s = sc.from_int(t_int)
            i = int(s)
            assert i == t_int

    @classmethod
    def t_from_list(cls, sc: type[sieve.Sievish]) -> None:
        vectors: list[tuple[list[int], int]] = [
            (cls.primes100, 100),
            (cls.primes30, 30),
        ]
        for liszt, n in vectors:
            s = sc.from_list(liszt, size=n)
            assert s.n == n
            assert liszt == list(s.primes())

    @classmethod
    def t_nomut(cls, sc: type[sieve.Sievish]) -> None:
        p30_in = cls.primes30.copy()
        p100_in = cls.primes100.copy()
        _ = sc.from_list(cls.primes30)
        _ = sc.from_list(cls.primes100)
        assert cls.primes100 == p100_in
        assert cls.primes30 == p30_in


class TestBaSieve:
    s_class = sieve.BaSieve

    def test_30(self) -> None:
        assert issubclass(self.s_class, sieve.BaSieve)
        Fixed.t_30(self.s_class)

    def test_count(self) -> None:
        Fixed.t_count(self.s_class)

    def test_primes(self) -> None:
        Fixed.t_primes(self.s_class)

    def test_2int(self) -> None:
        Fixed.t_2int(self.s_class)

    def test_from_int(self) -> None:
        Fixed.t_from_int(self.s_class)

    def test_from_list(self) -> None:
        Fixed.t_from_list(self.s_class)

    def test_test_nomut(self) -> None:
        Fixed.t_nomut(self.s_class)


class TestSetSieve:
    s_class = sieve.SetSieve

    def test_30(self) -> None:
        Fixed.t_30(self.s_class)

    def test_count(self) -> None:
        Fixed.t_count(self.s_class)

    def test_primes(self) -> None:
        Fixed.t_primes(self.s_class)

    def test_2int(self) -> None:
        Fixed.t_2int(self.s_class)

    def test_from_int(self) -> None:
        Fixed.t_from_int(self.s_class)

    def test_from_list(self) -> None:
        Fixed.t_from_list(self.s_class)

    def test_test_nomut(self) -> None:
        Fixed.t_nomut(self.s_class)


class TestIntSieve:
    s_class = sieve.IntSieve

    def test_30(self) -> None:
        Fixed.t_30(self.s_class)

    def test_count(self) -> None:
        Fixed.t_count(self.s_class)

    def test_primes(self) -> None:
        Fixed.t_primes(self.s_class)

    def test_2int(self) -> None:
        Fixed.t_2int(self.s_class)

    def test_from_int(self) -> None:
        Fixed.t_from_int(self.s_class)

    def test_from_list(self) -> None:
        Fixed.t_from_list(self.s_class)

    def test_test_nomut(self) -> None:
        Fixed.t_nomut(self.s_class)


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



from dataclasses import dataclass
import hashlib
from hmac import compare_digest
import math
import secrets
from typing import Callable
from toy_crypto import utils
from toy_crypto.nt import lcm, modinv

_DEFAULT_E = 65537


def default_e() -> int:
    """Returns the default public exponent, 65537"""
    return _DEFAULT_E


class DecryptionError(Exception):
    """
    For secure implementations it is important to not
    report why a decryption failed.

    See :func:`Oaep.allow_unsafe_messages` to enable
    reasons why decryption fails.
    """


type HashFunc = Callable[[bytes], hashlib._Hash]
"""Type for hashlib style hash function."""

type MgfFunc = Callable[[bytes, int, str], bytes]
"""Type for RFC8017 Mask Generation Function."""


class Oaep:
    """
    Tools and data for OAEP.

    Although this attempts to follow :rfc:`8017` in many
    respects, this is not designed to be interoperable
    with compliant keys and ciphertext.
    """

    _unsafe_messages: str = "forbid"
    """Anything other than "allow" will keep DecryptionError messages safe."""

    @classmethod
    def allow_unsafe_messages(cls, allow: bool = True) -> None:
        """Allow (or disallow) verbose DecryptionError messages."""

        cls._unsafe_messages = "allow" if allow else "forbid"

    @classmethod
    def are_unsafe_messages_allowed(cls) -> bool:
        """Does what it says on the tin."""

        if cls._unsafe_messages == "allow":
            return True
        return False

    @classmethod
    def _unsafe_msg(cls, message: str) -> str:
        if cls._unsafe_messages == "allow":
            return message
        return "Not telling, nohow!"

    @dataclass(frozen=True, kw_only=True)
    class HashInfo:
        """Information about hash function

        :param hashlib_name: Name as known by hashlib
        :param function: The callable function
        :param digest_size: in bytes
        :param input_limit: in bytes

        Note that names and identifiers here do not
        conform to RFCs. These are not mean to be interoperable
        with anything out in the world.
        """

        #: Name as known by hashlib
        hashlib_name: str
        function: HashFunc  #: The callable function itself
        digest_size: int  #: in bytes
        input_limit: int  #: maximum input, in bytes

    @dataclass(frozen=True, kw_only=True)
    class MgfInfo:
        """Information about Mask Generation function.

        :param algorithm: Name of the algorithm.
        :param hashAlgorithm: Key in :data:`KNOWN_HASHES`.
        :param function: A Callable mask generation function
        """

        algorithm: str  # eg "id-mfg1"
        hashAlgorithm: str  # Key in KNOWN_HASHES
        function: MgfFunc

    @staticmethod
    def mgf1(
        seed: bytes,  # There do not appear to be any constraints on this
        length: int,  # Output length
        hash_id: str,  # key for KNOWN_HASHES
    ) -> bytes:
        """Mask generation function.

        Generates a unique mask of length :data:`length` as described in
        :rfc:`appendix B.2.1 <8017#appendix-B.2.1>` of :rfc:`8017`.

        :param seed: This should come from a CSPRNG
        :param length: Length in bytes of the mask to generate.
        :param hash_id: The name hash function in :data:`KNOWN_HASHES`.

        :raises ValueError: if :data:`length` :math:`> 2^{32}` bytes.
        :raises ValueError: if :data:`hash_id` is unknown.
        """

        """
        I am using Pythonic variable naming instead of what is in the RFC

        RFC Name | My name | Description                        | Type
        -----------------------------------------------------------------
        mfgSeed | seed      | seed from which mask is generated | bytes
        maskLen | length    | Intended length of mask           | int
        mask    | mask      | Output                            | bytes
        T       | t         | Internal array for building mask  | bytearray
        C       | counter   | Counter, four octets              | bytes
                | hash_id   | ID of hash function               | str
        Hash    | hasher    | CS hash function                  | HashFunc
        """

        if length > 2**32:
            raise ValueError("mask too long")

        try:
            hash = Oaep.KNOWN_HASHES[hash_id]
        except KeyError:
            raise ValueError(f'Unsupported hash function: "{hash_id}')

        digest_size = hash.digest_size
        hasher = hash.function

        t = bytearray()

        # "For counter from 0 to \ceil (maskLen / hLen) - 1, ..."
        # range is not inclusive at high end. So we need to add one
        for c in range(math.ceil(length / digest_size) - 1 + 1):
            counter = Oaep.i2osp(c, 4)
            assert len(counter) == 4
            digest = hasher(seed + counter).digest()
            t.extend(digest)

        assert len(t) >= length
        mask = bytes(t[:length])
        return mask

    KNOWN_HASHES: dict[str, HashInfo] = {
        "sha256": HashInfo(
            hashlib_name="sha256",
            function=hashlib.sha256,
            digest_size=32,
            input_limit=2**61 - 1,
        ),
        # We need sha1 because that is what test vectors exist for
        "sha1": HashInfo(
            hashlib_name="sha1",
            function=hashlib.sha1,
            digest_size=20,
            input_limit=2**64 - 1,
        ),
    }
    """Hashes known for OAEP. keys will be hashlib names."""

    KNOWN_MGFS: dict[str, MgfInfo] = {
        "mgf1SHA256": MgfInfo(
            algorithm="id_mgf1", hashAlgorithm="sha256", function=mgf1
        ),
        "mgf1SHA1": MgfInfo(
            algorithm="id_mgf1", hashAlgorithm="sha1", function=mgf1
        ),
    }
    """Known Mask Generation Functions."""

    @staticmethod
    def i2osp(n: int, length: int) -> bytes:
        """Integer to an octet string of length length.

        Implements function from :rfc:`RFC 8017 §4.1 <8017#section-4.1>`.

        :param n: A non-negative integer
        :param length: Length of returned bytes object

        :raises ValueError: if :data:`n` is negative.
        :raises ValueError: if :data:`n` cannot fit in :data:`length` bytes

        .. warning::

            When called from a decryption operation, exceptions should be caught
            and handled discretely.


        All operations big-endian.
        """

        if n < 0:
            raise ValueError("Number cannot be negative")

        if (n.bit_length() + 7) // 8 > length:
            raise ValueError("input is too large for the given length")

        return n.to_bytes(length, byteorder="big", signed=False)

    @staticmethod
    def os2ip(x: bytes) -> int:
        """octet-stream to unsigned big-endian int.

         Implements function from :rfc:`RFC 8017 §4.2 <8017#section-4.2>`.

        :param x:
            The octet-stream (:py:class:`bytes`) you want
            to make an :py:class:`int` from.

        Returned is a non-negative integer.

        All operations are big-endian.
        """

        return int.from_bytes(x, byteorder="big", signed=False)


class PublicKey:
    def __init__(self, modulus: int, public_exponent: int) -> None:
        """Public key from public values."""
        self._N = modulus
        self._e = public_exponent

    @property
    def N(self) -> int:
        """Public modulus N."""
        return self._N

    @property
    def e(self) -> int:
        """Public exponent e"""
        return self._e

    def encrypt(self, message: int) -> int:
        """Primitive encryption with neither padding nor nonce.

        :raises ValueError: if message < 0
        :raises ValueError: if message isn't less than the public modulus
        """

        if message < 0:
            raise ValueError("Positive messages only")

        """
        There is a reason for the explicit conversion to int in the
        comparison below. If message was created as a member of a SageMath
        finite group mod N, self._N would be converted to that before
        comparison and self._N ≡ 0 (mod self._N).
        """
        if not int(message) < self._N:
            raise ValueError("Message too big")

        return pow(base=message, exp=self._e, mod=self._N)

    def oaep_encrypt(
        self,
        message: bytes,
        label: bytes = b"",
        hash_id: str = "sha256",
        mgf_id: str = "mgf1SHA256",
        _seed: bytes | None = None,  # For testing only
    ) -> bytes:
        """
        RSA OAEP encryption.

        :param message: The message to encrypt.
        :param label: Rarely used. Just leave as default.
        :param hash_id: Name of the hash function.
        :param mgf_id: Name of the MGF function (with hash).
        :param _seed:
            Used for testing only. OAEP is not supposed to be deterministic.

        :raises ValueError: if hash or MGF is not recognized.
        :raises ValueError:
            if lengths of inputs exceed what modulus and hash sizes
            can accommodate.

        https://datatracker.ietf.org/doc/html/rfc8017#section-7.1.1
        """

        try:
            h = Oaep.KNOWN_HASHES[hash_id]
        except KeyError:
            raise ValueError(f'Unsupported hash: "{hash_id}')

        try:
            mgf = Oaep.KNOWN_MGFS[mgf_id]
        except KeyError:
            raise ValueError(
                f'Unsupported mask generation function: "{mgf_id}'
            )

        if len(label) > h.input_limit:
            raise ValueError("label too long")

        k = (self.N.bit_length() + 7) // 8  # length of N in bytes

        if len(message) > k - 2 * h.digest_size - 2:
            raise ValueError("message too long")

        lhash = h.function(label).digest()

        ps_length = k - len(message) - 2 * h.digest_size - 2
        padding_string = bytes(ps_length)

        data_block = lhash + padding_string + bytes([0x01]) + message

        # So that we can test with a set seed
        seed: bytes
        if _seed is None:
            seed = secrets.token_bytes(h.digest_size)
        else:
            seed = _seed

        mask = mgf.function(seed, k - h.digest_size - 1, mgf.hashAlgorithm)
        masked_db = utils.xor(data_block, mask)
        seed_mask = mgf.function(masked_db, h.digest_size, mgf.hashAlgorithm)
        masked_seed = utils.xor(seed, seed_mask)

        encoded_m = bytes([0x00]) + masked_seed + masked_db
        m = Oaep.os2ip(encoded_m)
        ctext = self.encrypt(m)

        return Oaep.i2osp(ctext, k)

    def __eq__(self, other: object) -> bool:
        """True when each has the same modulus and public exponent.

        When comparing to a PrivateKey, this compares only the public parts.
        """
        if isinstance(other, PublicKey):
            return self.e == other.e and self.N == other.N

        return NotImplemented


class PrivateKey:
    def __init__(self, p: int, q: int, pub_exponent: int = _DEFAULT_E) -> None:
        """RSA private key from primes p and q.

        This does not perform any sanity checks on p and q.
        It is your responsibility to ensure that p and q are prime

        :raises ValueError: if e is not coprime with lcm(p - 1, q - 1).
        """

        self._p = p
        self._q = q
        self._e = pub_exponent

        self._N = self._p * self._q
        self._pubkey = PublicKey(self._N, self._e)

        self._dP = modinv(self._e, p - 1)
        self._dQ = modinv(self._e, (self._q - 1))
        self._qInv = modinv(self._q, self._p)

        try:
            self._d = self._compute_d()
        except ValueError:
            raise ValueError("p, q, and e are incompatible with each other ")

    @property
    def pub_key(self) -> PublicKey:
        """The public key corresponding to self.

        The public key does not contain any secrets.
        """

        return self._pubkey

    @property
    def e(self) -> int:
        """Public exponent."""
        return self._e

    def __eq__(self, other: object) -> bool:
        """True iff keys are mathematically equivalent

        Private keys with internal differences can behave identically
        with respect to input and output. This comparison will return
        True when they are equivalent in this respect.

        When compared to a PublicKey, this compares only the public part.
        """
        if isinstance(other, PrivateKey):
            return self.pub_key == other.pub_key

        if isinstance(other, PublicKey):
            return self.pub_key == other

        return NotImplemented

    def _compute_d(self) -> int:
        λ = lcm(self._p - 1, self._q - 1)
        try:
            return modinv(self.e, λ)
        except ValueError:
            raise ValueError("Inverse of e mod λ does not exist")

    def decrypt(self, ciphertext: int) -> int:
        """Primitive decryption.

        :param ciphertext: Ciphertext as :py:class:`int`
        :raises ValueError: if :data:`ciphertext` is out of range for this key.

        """

        ciphertext = int(ciphertext)  # See comment in PublicKey.encrypt()

        if ciphertext < 1 or ciphertext >= self.pub_key.N:
            raise ValueError("ciphertext is out of range")

        # m =  pow(base=ciphertext, exp=self._d, mod=self._N)
        # but we will use the CRT
        # version comes from  rfc8017 §5.1.2

        m_1 = pow(ciphertext, self._dP, self._p)
        m_2 = pow(ciphertext, self._dQ, self._q)

        # CRT for the win!
        h = ((m_1 - m_2) * self._qInv) % self._p

        m = m_2 + self._q * h
        return m

    def oaep_decrypt(
        self,
        ciphertext: bytes,
        label: bytes = b"",
        hash_id: str = "sha256",
        mgf_id: str = "mgf1SHA256",
    ) -> bytes:
        """
        RSA OAEP decryption.

        :param ciphertext: The message to encrypt.
        :param label: Rarely used.
        :param hash_id: Name of the hash function.
        :param mgf_id: Name of the MGF function (with hash).

        :raises ValueError: if hash or MGF is not recognized.
        :raises DecryptionError:
            on various decryption errors.
            If unsafe error reporting is enabled, details of
            decryption errors will be provided.

        """

        try:
            h = Oaep.KNOWN_HASHES[hash_id]
        except KeyError:
            raise ValueError(f'Unsupported hash: "{hash_id}')

        try:
            mgf = Oaep.KNOWN_MGFS[mgf_id]
        except KeyError:
            raise ValueError(
                f'Unsupported mask generation function: "{mgf_id}'
            )
        k = (
            self.pub_key.N.bit_length() + 7
        ) // 8  # ceil(length of N in bytes)

        # checks in Step 1 involve no secrets
        # Step 1.a
        if len(label) > h.input_limit:
            raise DecryptionError(Oaep._unsafe_msg("Label too long"))

        # Step 1.b
        if len(ciphertext) != k:
            raise DecryptionError(Oaep._unsafe_msg("Ciphertext is wrong size"))

        # Step 1.c
        if k < 2 * h.digest_size + 2:
            raise DecryptionError(Oaep._unsafe_msg("Modulus is way too short"))

        # Step 2.a
        c = Oaep.os2ip(ciphertext)

        # Step 2.b
        try:
            m = self.decrypt(c)
        except Exception as e:
            raise DecryptionError(
                Oaep._unsafe_msg(f"Primitive decryption error: {e}")
            )

        # Step 2.c
        # if k is computed correctly, conversion of bytes
        # should never fail.
        em = Oaep.i2osp(m, k)

        # Step 3 is where we perform validations that make use of secrets

        # Step 3.a is done before other tests to thwart certain timing attacks
        lhash = h.function(label).digest()

        # Step 3.b. Parse encoded message
        y: int = em[0]
        masked_seed: bytes = em[1 : h.digest_size + 1]
        masked_datablock: bytes = em[h.digest_size + 1 :]

        # Step 3.c and 3.d
        seed_mask = mgf.function(
            masked_datablock, h.digest_size, mgf.hashAlgorithm
        )
        seed = utils.xor(masked_seed, seed_mask)

        # Steps 3.e and 3.f
        db_mask = mgf.function(seed, k - h.digest_size - 1, mgf.hashAlgorithm)
        data_block = utils.xor(masked_datablock, db_mask)

        # Parsing portion of Step 3.g
        lhash_prime: bytes = data_block[: h.digest_size]
        remainder: bytes = data_block[h.digest_size :].lstrip(bytes([0]))
        one: int = remainder[0]
        message: bytes = remainder[1:]

        # Validation portion of Step 3.g
        if one != 1:
            raise DecryptionError(Oaep._unsafe_msg("Expected 0x01"))
        if not compare_digest(lhash, lhash_prime):
            raise DecryptionError(Oaep._unsafe_msg("Label mismatch"))
        if y != 0:
            raise Exception(Oaep._unsafe_msg("Expected 0x00 leading byte"))

        return message



from typing import Any
import pydot  # types: ignore
from toy_crypto.sec_games import Ind, IndEav, IndCpa, IndCca1, IndCca2
from toy_crypto.sec_games import State


games: dict[str, object] = {
    "IND-EAV": IndEav,
    "IND-CPA": IndCpa,
    "IND-CCA1": IndCca1,
    "IND-CCA2": IndCca2,
}

extra_state_abbr: list[str] = list("αβγδεζηθικλμνξπρστφχψω")[::-1]


def make_graph(game: Ind[Any], name: str) -> pydot.Graph:
    ttable = game.T_TABLE
    states: list[State] = list(ttable.keys())
    label = f"State transitions in {name} game"
    graph = pydot.Dot(
        name,
        graph_type="digraph",
        rankdir="LR",
        ranksep=1.5,
        nodesep=0.75,
    )
    graph.set_node_defaults(shape="circle")
    graph.set_edge_defaults(
        penwidth=0.75,
        labelfloat=True,
        fontname="Anonymous Pro Bold",
        fontsize=12,
    )

    for state in states:
        for label, destination in ttable[state].items():
            label = label + "()"
            edge = pydot.Edge(state, destination, label=label)
            graph.add_edge(edge)

    return graph


for name, game in games.items():
    graph: pydot.Graph = make_graph(game, name)  # type: ignore[arg-type]
    graph.write_raw(f"{name}.gv")



"""
Various utilities for manipulationg bit-like things.

The utilities here are most subject to change
as many where just quick things I needed.
This is more of a trash heap than a well-thought-out module.
"""

from typing import Optional, Any, Union, Callable, Self
from collections.abc import Iterator
import operator
from .utils import xor
from .types import SupportsBool


def bits(n: int) -> Iterator[int]:
    """
    0s and 1s representing bits of n, starting with the least significant bit.

    :raises ValueError: if n is negative.
    """

    n = int(n)
    if n < 0:
        raise ValueError("n cannot be negative")

    while n > 0:
        yield n & 1
        n >>= 1


def get_bit(n: int, k: int) -> int:
    """returns k-th bit of n."""

    if k < 0:
        raise ValueError("k cannot be negative")
    return (n & (1 << (k))) >> (k)


def set_bit(n: int, k: int, value: bool | int = True) -> int:
    """Returns copy of n with k-th bit set to value."""

    # k must be greater than 0
    if k < 0:
        raise ValueError("k cannot be negative")

    if not value:
        # setting to 0
        return n & ~(1 << k)

    return (1 << k) | n


def bit_index_linear(n: int, k: int, b: bool | int = 1) -> int | None:
    """Returns which bit is the k-th b bit in n.

    This is to mimic bitarray.utils.count_n, but for working with python
    built-in int
    """

    if k < 1:
        raise ValueError("n must be positive")

    bc = n.bit_count() if b else (~n).bit_count()

    if k > bc:
        return None
    b = 1 if b else 0

    # naive code that just does a linear search count through bits of n
    count = 0
    for i in range(n.bit_length()):
        n, r = divmod(n, 2)
        if r == b:
            count += 1
            if count >= k:
                return i
    return None  # This really shouldn't ever be reached.


def bit_index(n: int, k: int, b: bool | int = 1) -> int | None:
    """Returns which bit is the k-th b bit in n.

    This is to mimic bitarray.utils.count_n, but for working with python
    built-in int
    """

    if k < 1:
        raise ValueError("k must be positive")

    bc = n.bit_count() if b else (~n).bit_count()
    bl = n.bit_length()

    if k > bc:
        return None
    b = 1 if b else 0

    linear_theshold = 31
    if bc <= linear_theshold:
        return bit_index_linear(n, k, b)

    mid_index = bl // 2
    midpoint = int(2**mid_index)
    upper, lower = divmod(n, midpoint)

    # Debugging check against alternative way to compute upper and lower
    assert upper == n >> mid_index
    assert lower == n & ((1 << mid_index) - 1)

    bcl = lower.bit_count()
    if bcl == k:
        return mid_index - 1
    if bcl > k:
        ret_val = bit_index(lower, k, b)
        return ret_val

    u_count = bit_index(upper, k - bcl, b)
    if u_count is not None:
        return bcl + u_count

    return None


class Bit:
    """
    Because I made poor choices earlier of how to represent
    bits, I need an abstraction.
    """

    def __init__(self, b: SupportsBool) -> None:
        self._value: bool = b is True
        self._as_int: Optional[int] = None
        self._as_bytes: Optional[bytes] = None

    def __bool__(self) -> bool:
        """Every Bit should be Truthy."""
        return self._value

    def as_int(self) -> int:
        """Returns 0 or 1."""
        if not self._as_int:
            self._as_int = 1 if self._value else 0
        return self._as_int

    def as_bytes(self) -> bytes:
        """Returns a big-endian byte. Either 0x00 or 0x01."""
        if not self._as_bytes:
            self._as_bytes = (
                (1).to_bytes(1, "big")
                if self._value
                else (0).to_bytes(1, "big")
            )
        return self._as_bytes

    def __eq__(self, other: Any) -> bool:
        """Equality with other."""
        ob = self._other_bool(other)
        if ob is None:
            return NotImplemented
        ob = bool(other)

        return self._value == ob

    @staticmethod
    def _other_bool(other: Any) -> Optional[bool]:
        if isinstance(other, bytes):
            ob = any([b != 0 for b in other])
        elif not isinstance(other, SupportsBool):
            return None
        else:
            ob = other.__bool__()
        return ob

    def _logic(
        self, other: bool, expr: Callable[[bool, bool], bool]
    ) -> Union["Bit", int, bool, bytes]:
        """
        Abstraction to manage type of :data:`other`
        for things like :func:`__and__` and :func:`__or__`.
        """
        sb = bool(self)
        tvalue = expr(sb, other)

        if isinstance(other, Bit):
            return Bit(tvalue)

        if isinstance(other, int):
            return 1 if tvalue else 0
        if isinstance(other, bytes):
            return (1).to_bytes(1, "big") if tvalue else (0).to_bytes(1, "big")

        return tvalue

    def __and__(self, other: Any) -> Union["Bit", int, bool, bytes]:
        """For the "&" operator."""
        ob = self._other_bool(other)
        if ob is None:
            return NotImplemented

        return self._logic(other=ob, expr=lambda s, o: s and o)

    def __xor__(self, other: Any) -> Union["Bit", int, bool, bytes]:
        """For the "^" operator."""
        ob = self._other_bool(other)
        if ob is None:
            return NotImplemented

        return self._logic(other=ob, expr=lambda s, o: operator.xor(s, o))

    def __or__(self, other: Any) -> Union["Bit", int, bool, bytes]:
        """For the "|" operator."""
        ob = self._other_bool(other)
        if ob is None:
            return NotImplemented

        return self._logic(other=ob, expr=lambda s, o: s or o)

    def inv(self) -> "Bit":
        inv_b = not self
        return Bit(inv_b)

    def __inv__(self) -> "Bit":
        return self.inv()


def set_bit_in_byte(byte: int, bit: int, value: SupportsBool) -> int:
    """Sets the bit-most significant bit to value in byte."""
    byte %= 256
    bit %= 8

    if value:
        byte |= 1 << bit
    else:
        byte &= ~(1 << bit)
    return byte % 256


def flip_end(byte: int) -> int:
    """Return int with reversed bit sequence"""
    if byte < 0 or byte > 255:
        raise ValueError("byte is not representable as byte")
    result = 0
    for p in range(8):
        byte, b = divmod(byte, 2)
        result += b * (2 ** (8 - p))
    return result


def hamming_distance(a: bytes, b: bytes) -> int:
    """Hamming distance between byte sequences of equal length.

    :raises ValueError: if len(a) != len(b).
    """

    if len(a) != len(b):
        raise ValueError("Lengths are unequal")

    # hamming distance will be the number of 1 bits in a xor b
    db: bytes = xor(a, b)
    # bit_count is only defined for ints, so
    return int.from_bytes(db, signed=False).bit_count()


class PyBitArray:
    """A pure Python bitarray-like object.

    This does not implement all methods of bitarray,
    nor does it fully follow the bitarray API.

    This is very much a work in progress.
    """

    """
    Internal little-endian bytearray of big endian bytes,
    but that shouldn't be seen by users.

    So the bit at index 0 will be the right-most bit of the left-most byte.
    If I do this right, users will never have to know or deal with that.
    """

    def __init__(self, bit_length: int, fill_bit: SupportsBool = 0) -> None:
        # Instance attributes that should always exist
        self._data: bytearray
        self._length: int  # length in used bits
        self._free_bits: int  # number of unused bits in last byte

        if bit_length < 0:
            raise ValueError("bit_length cannot be negative")

        fill_byte: int
        if not fill_bit:
            fill_byte = 0
        else:
            fill_byte = 255

        self._length = bit_length
        byte_len, self._free_bits = divmod(self._length, 8)
        if self._free_bits > 0:
            byte_len += 1
        self._data = bytearray([fill_byte] * byte_len)
        self._data[-1] >>= self._free_bits

    @classmethod
    def from_int(cls, n: int) -> Self:
        """New instance from int."""
        instance = cls(n.bit_length())

        idx = 0
        while n:
            n, r = divmod(n, 2)
            if r:
                instance[idx] = 1
            idx += 1

        return instance

    def append(self, b: SupportsBool) -> None:
        """appends bit b."""
        b = 1 if b else 0

        self._length += 1
        if self._free_bits == 0:
            self._data.append(b)
        else:
            self[-1] = b

        self._free_bits -= 1
        self._free_bits %= 8

    def _inner_getitem(self, index: int) -> int:
        while index < 0:
            index += self._length
        byte_index, bit_index = divmod(index, 8)
        byte = self._data[byte_index]
        value = byte & (1 << bit_index)
        return 1 if value != 0 else 0

    def __getitem__(self, index: int) -> int:
        """Retrieve a bit using [] notation."""
        if not index < self._length:
            raise IndexError
        return self._inner_getitem(index)

    def __setitem__(self, index: int, value: SupportsBool) -> None:
        """Set a bit using [] notation."""
        while index < 0:
            index += self._length

        byte_index, bit_index = divmod(index, 8)
        byte = self._data[byte_index]

        new_byte = set_bit_in_byte(byte, bit_index, value)
        self._data[byte_index] = new_byte

    @property
    def nbytes(self) -> int:
        """Length in bytes"""
        return len(self._data)

    @property
    def padbits(self) -> int:
        """Number of pad bits"""
        return self._free_bits

    def __len__(self) -> int:
        """len() is defined."""
        return self._length

    def bits(
        self, start: int = 0, stop: int | None = None, step: int = 1
    ) -> Iterator[int]:
        """Iterator of bits, from least significant."""
        if step == 0:
            raise ValueError("step cannot be 0")
        step_sign = 1 if step > 0 else -1

        match (stop, step_sign):
            case None, 1:
                stop = self._length
            case None, _:  # sign is negative
                stop = 0
            case _, _:
                stop = stop

        # I know this, you know this. Now the type checker will know this
        assert stop is not None
        indices = range(start, stop, step)

        for idx in indices:
            yield self[idx]

    def __iter__(self) -> Iterator[int]:
        return self.bits()

    def __int__(self) -> int:
        """Integer with first element as least significant bit"""
        return sum(b * (2**i) for i, b in enumerate(self))

    def count(self) -> int:
        """Returns number of 1 bits"""
        return int.from_bytes(self._data).bit_count()

    @classmethod
    def from_bytes(cls, byte_data: bytes, endian: str = "big") -> Self:
        ints: list[int]
        match endian.lower():
            case "big":
                ints = [int(b) for b in byte_data]
            case "little":
                ints = [int(b) for b in reversed(byte_data)]
            case _:
                raise ValueError('endian must be "big" or "little"')

        instance = cls(8 * len(ints))
        instance._data = bytearray(ints)

        return instance

    def to_bytes(self, endian: str = "big") -> bytes:
        if endian not in ["big", "little"]:
            raise ValueError('endian must be "big" or "little"')

        if endian == "big":
            return bytes(self._data)
        return bytes(self._data[::-1])

    def bit_index(self, k: int, b: int = 1) -> int | None:
        """The index of the n-th 1 or 0 bit."""

        # For the moment I will just use bit_index for int
        # TODO: Write b-search for bit index using native data
        as_int = int(self)
        return bit_index(as_int, k, b)



import secrets
import sys

import pytest
from toy_crypto.sec_games import IndEav, IndCpa, StateError, IndCca2


class TestInd:
    """
    For reliable tests, we will need a encryption scheme that the adversary
    can win against all the time.
    """

    def encryptor(self, key: int, m: bytes) -> bytes:
        encrypted: bytes = bytes([(b + key) % 256 for b in m])
        return encrypted

    def decryptor(self, key: int, m: bytes) -> bytes:
        encrypted: bytes = bytes([(b - key) % 256 for b in m])
        return encrypted

    def key_gen(self) -> int:
        return secrets.randbelow(256)

    def test_eav(self) -> None:
        challenger = IndEav(self.key_gen, self.encryptor)

        m0 = b"AA"
        m1 = b"AB"

        # chance of false pass is 2^trials
        trials = 20
        for _ in range(trials):
            challenger.initialize()
            ctext = challenger.encrypt_one(m0, m1)
            guess: bool = ctext[0] != ctext[1]

            assert challenger.finalize(guess)

    def test_eav_once(self) -> None:
        challenger = IndEav(self.key_gen, self.encryptor)
        m0 = b"AA"
        m1 = b"AB"

        challenger.initialize()

        _ = challenger.encrypt_one(m0, m1)

        with pytest.raises(StateError):
            _ = challenger.encrypt_one(m0, m1)

    def test_cpa_deterministic(self) -> None:
        """Wins against deterministic encryption"""
        challenger = IndCpa(self.key_gen, self.encryptor)

        m0 = b"Attack at dawn!"
        m1 = b"Attack at dusk!"

        # chance of false pass is 2^trials
        trials = 20
        for _ in range(trials):
            challenger.initialize()
            ctext1 = challenger.encrypt_one(m0, m1)
            ctext2 = challenger.encrypt_one(m1, m1)
            guess = ctext1 == ctext2

            assert challenger.finalize(guess)

        """Wins against deterministic encryption"""
        challenger = IndCpa(self.key_gen, self.encryptor)

        m0 = b"Attack at dawn!"
        m1 = b"Attack at dusk!"

        # chance of false pass is 2^trials
        trials = 20
        for _ in range(trials):
            challenger.initialize()
            ctext1 = challenger.encrypt_one(m0, m1)
            ctext2 = challenger.encrypt_one(m1, m1)
            guess = ctext1 == ctext2

            assert challenger.finalize(guess)

    def test_cca2(self) -> None:
        """Wins against deterministic encryption.

        I should construct a test that distinguishes CPA and CCA,
        but this is not that test
        """
        challenger = IndCca2(
            key_gen=self.key_gen,
            encryptor=self.encryptor,
            decrytpor=self.decryptor,
        )

        m0 = b"Attack at dawn!"
        m1 = b"Defend at dusk!"
        cct1 = b"AAAAAAA"

        # chance of false pass is 2^trials
        trials = 20
        for _ in range(trials):
            challenger.initialize()
            c0 = challenger.encrypt(m0)
            ctext = challenger.encrypt_one(m0, m1)

            if c0 == ctext:
                first_guess = 0
            else:
                first_guess = 1

            # let's use a chosen ciphertext to break this
            # even though it could be broken otherwise.
            ptext = challenger.decrypt(cct1)
            key_guess = cct1[0] - ptext[0] % 256
            if (m1[0] + key_guess) % 256 == ctext[0]:
                second_guess = 1
            else:
                second_guess = 0

            assert first_guess == second_guess

            assert challenger.finalize(first_guess)

    def test_cca1(self) -> None:
        """Wins against deterministic encryption.

        I should construct a test that distinguishes CPA and CCA,
        but this is not that test
        """
        challenger = IndCca2(
            key_gen=self.key_gen,
            encryptor=self.encryptor,
            decrytpor=self.decryptor,
        )

        m0 = b"Attack at dawn!"
        m1 = b"Defend at dusk!"
        cct1 = b"AAAAAAA"

        # chance of false pass is 2^trials
        trials = 20
        for _ in range(trials):
            challenger.initialize()

            # let's use a chosen ciphertext to break this
            # even though it could be broken otherwise.
            ptext = challenger.decrypt(cct1)
            key_guess = cct1[0] - ptext[0] % 256

            ctext = challenger.encrypt_one(m0, m1)

            if (m1[0] + key_guess) % 256 == ctext[0]:
                first_guess = 1
            else:
                first_guess = 0

            c0 = challenger.encrypt(m0)

            if c0 == ctext:
                second_guess = 0
            else:
                second_guess = 1

            assert first_guess == second_guess

            assert challenger.finalize(first_guess)

    def test_cca2_ctext_clear(self) -> None:
        """Previous round banned ctext allowed.

        This tests a bug in which the list of banned ciphertexts that
        can be given to decrypt() wasn't clear with re-initialization.
        """
        challenger = IndCca2(
            key_gen=self.key_gen,
            encryptor=self.encryptor,
            decrytpor=self.decryptor,
        )

        m0 = b"Attack at dawn!"
        m1 = b"Defend at dusk!"
        x = b"B"
        cct1 = b"AAAAAAA"

        challenger.initialize()
        _ = challenger.decrypt(cct1)
        ctext = challenger.encrypt_one(m0, m1)

        with pytest.raises(Exception):
            _ = challenger.decrypt(ctext)

        # check that both encrypt and decrypt can still be called
        c_x = challenger.encrypt(x)
        p_x = challenger.decrypt(c_x)
        assert p_x == x
        _ = challenger.finalize(0)

        # Now the thing we really aim to test here.
        challenger.initialize()
        _ = challenger.decrypt(c_x)

        new_ctext = challenger.encrypt_one(b"X", b"Y")

        # This call should be allowed
        _ = challenger.decrypt(ctext)

        with pytest.raises(Exception):
            _ = challenger.decrypt(new_ctext)


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
import tomllib
from docutils.parsers.rst import Directive

from pprint import pformat
from importlib import import_module
from docutils import nodes
from sphinx import addnodes
from sphinx.application import Sphinx

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../../src"))

import toy_crypto  # noqa

from toy_crypto import __about__  # noqa: E402

version = __about__.__version__

# Pull general sphinx project info from pyproject.toml
# Modified from https://stackoverflow.com/a/75396624/1304076
with open("../../pyproject.toml", "rb") as f:
    toml = tomllib.load(f)

pyproject = toml["project"]

project = pyproject["name"]
release = version
author = ",".join([author["name"] for author in pyproject["authors"]])
copyright = f"2024–2025 {author}"

github_username = "jpgoldberg"
github_repository = "toy-crypto-math"


# From
# https://github.com/sphinx-doc/sphinx/issues/11548#issuecomment-1693689611


class PrettyPrintIterable(Directive):
    """
    Definition of a custom directive to pretty-print an iterable object.

    This is used in place of the automatic API documentation
    only for module variables which would just print a long signature.
    """

    required_arguments = 1

    def run(self):  # type: ignore
        paths = self.arguments[0].rsplit(".", 2)
        module_path = paths[0]
        module = import_module(module_path)
        member = getattr(module, paths[1])
        if len(paths) == 3:
            member = getattr(member, paths[2])

        code = pformat(
            member,
            indent=2,
            width=80,
            depth=3,
            compact=False,
            sort_dicts=False,
        )

        literal = nodes.literal_block(code, code)
        literal["language"] = "python"

        return [addnodes.desc_content("", literal)]


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions: list[str] = [
    "sphinx_toolbox.more_autodoc.augment_defaults",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
    "enum_tools.autoenum",
    "sphinx_toolbox.github",
    "sphinx_toolbox.decorators",
    "sphinx_toolbox.wikipedia",
    # "sphinx_toolbox.installation",
    # "sphinx_toolbox.more_autodoc.typehints",
    "sphinx_autodoc_typehints",
    "sphinx_toolbox.more_autodoc.autoprotocol",
    "sphinx_toolbox.more_autodoc.genericalias",
    "sphinx_toolbox.more_autodoc.typevars",
    # "sphinx_toolbox.more_autodoc.variables",
]

autodoc_typehints = "signature"
typehints_use_signature = True
typehints_use_signature_return = True
always_document_param_types = True
typehints_defaults = "comma"


autodoc_show_sourcelink = True

extensions.append("sphinx.ext.intersphinx")
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "dns": ("https://dnspython.readthedocs.io/en/stable/", None),
}

extensions.append("sphinxcontrib.bibtex")
bibtex_bibfiles = ["refs.bib"]
bibtex_reference_style = "author_year"


def setup(app: Sphinx) -> None:
    """Set up the final Sphinx application.

    This function loads any other customization that was added in this
    configuration file, thus making it itself a Sphinx extension.
    """
    app.add_directive("pprint", PrettyPrintIterable)


rst_prolog = f"""
.. |project| replace:: **{project}**
.. |root| replace:: :mod:`toy_crypto`
.. _pyca: https://cryptography.io/en/latest/
.. _SageMath: https://www.sagemath.org
.. _primefac: https://pypi.org/project/primefac/
.. _bitarray: https://github.com/ilanschnell/bitarray
.. _pypkcs1: https://github.com/bdauvergne/python-pkcs1/
.. _pypi: https://pypi.org/
"""


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "nature"
html_static_path = ["_static"]



"""
Helpful(?) type declarations and guards.

These are intended to make things easier for me, the author (jpgoldberg).
They are not carefully thought out.
This module is probably the least stable of any of these unstable modules.
"""

from typing import (
    Any,
    NewType,
    TypeGuard,
    Protocol,
    runtime_checkable,
)


Prob = NewType("Prob", float)
"""Probability: A float between 0.0 and 1.0"""


def is_prob(val: Any) -> TypeGuard[Prob]:
    """true if val is a float, s.t. 0.0 <= va <= 1.0"""
    if not isinstance(val, float):
        return False
    return val >= 0.0 and val <= 1.0


PositiveInt = NewType("PositiveInt", int)
"""Positive integer."""


def is_positive_int(val: Any) -> TypeGuard[PositiveInt]:
    """true if val is a float, s.t. 0.0 <= val <= 1.0"""
    if not isinstance(val, int):
        return False
    return val >= 1


Byte = int
"""And int representing a single byte.

Currently implemented as a type alias.
As a consequence, type checking is not going to identify
cases where an int out of the range of a byte is used.
"""


def is_byte(val: Any) -> bool:
    """True iff val is int s.t. 0 <= val < 256."""
    if not isinstance(val, int):
        return False
    return 0 <= val and val < 256


@runtime_checkable
class SupportsBool(Protocol):
    def __bool__(self) -> bool: ...



from collections.abc import Callable, KeysView, Mapping
from dataclasses import dataclass
from enum import StrEnum
import secrets
from typing import Any, Generic, Optional, TypeVar, cast, Protocol
from functools import wraps
from toy_crypto.types import SupportsBool
from toy_crypto.utils import hash_bytes

K = TypeVar("K")
"""Unbounded type variable intended for any type of key."""

type KeyGenerator[K] = Callable[[], K]
"""To describe key generation functions"""

type Cryptor[K] = Callable[[K, bytes], bytes]
"""To describe encryptor/decryptor functions."""


class StateError(Exception):
    """When something attempted in an inappropriate state."""


class State(StrEnum):
    """The state a game."""

    STARTED = "S"
    """Game has not been initialized."""

    INITIALIZED = "I"
    """Game is initialized"""

    CHALLENGED = "C"
    """Challenge text created."""


class Action(StrEnum):
    """Adversary actions (Methods called)."""

    INITIALIZE = "initialize"
    """initialize() called."""

    ENCRYPT_ONE = "encrypt_one"
    """encrypt_one() called."""

    ENCRYPT = "encrypt"
    """encrypt() called."""

    DECRYPT = "decrypt"
    """decrypt() called"""

    FINALIZE = "finalize"
    """finalize() called."""


@dataclass
class TransitionTable:
    """Transition Table to manage state of a game.

    It can be treated like a mapping.
    """

    table: Mapping[State, Mapping[Action, State]]

    def keys(self) -> KeysView[State]:
        """Just like ``keys`` for a real dict."""
        return self.table.keys()

    def __getitem__(self, item: State) -> Mapping[Action, State]:
        """So that items can be looked up with ``[key]`` as in a real dict."""
        return self.table[item]

    def __str__(self) -> str:
        pad = "  "
        if len(self.table) == 0:
            return "{ }"
        result = "{\n"
        for key in self.keys():
            result += f"{pad}State.{key.name}:\n"
            for act in self.table[key].keys():
                result += f"{pad * 2}Action.{act.name}"
                result += f" -> State.{self.table[key][act].name}\n"
        result += "}"
        return result


class SupportsTTable(Protocol):
    """Has what it takes to be decorated by :func:`manage_state`."""

    t_table: TransitionTable
    current_state: State
    _state: State


# This lexical scoping trickery is based on
#    https://stackoverflow.com/a/38286176/1304076
# with a hat tip to https://www.reddit.com/user/GeorgeFranklyMathnet/
#  https://www.reddit.com/r/learnpython/comments/1i34vgh/comment/m7k6hgn/
def manage_state[F: Callable[..., Any]](fn: F) -> F:
    """Decorator to check/transition state for Ind method calls."""
    action = Action(fn.__name__)

    @wraps(fn)
    def decorator(self: SupportsTTable, *args, **kwargs):  # type: ignore
        if action not in self.t_table[self.current_state]:
            raise StateError(
                f"{action} not allowed in state {self.current_state}"
            )
        retvalue = fn(self, *args, **kwargs)
        self._state = (self.t_table[self._state])[action]
        return retvalue

    return cast(F, decorator)


class Ind(Generic[K]):
    T_TABLE: TransitionTable

    #: Game does not track which challenge texts have been created
    TRACK_CHALLENGE_CTEXTS: bool

    def __init__(
        self,
        key_gen: KeyGenerator[K],
        encryptor: Cryptor[K],
        decryptor: Optional[Cryptor[K]] = None,
        transition_table: Optional[TransitionTable] = None,
    ) -> None:
        """
        A super class for symmetric Indistinguishability games.

        Unless the user provides an appropriate transition table,
        no methods will be allowed.
        """

        self._key_gen = key_gen
        self._encryptor = encryptor
        self._decryptor = decryptor if decryptor else self._undefined_decryptor

        self._key: Optional[K] = None
        self._b: Optional[bool] = None
        self._state = State.STARTED

        self._challenge_ctexts: set[str] = set()

        """
        Each state is a dictionary of [Transition : State_Name]
        Transitions are the names of methods (or "start")
        """

        self._t_table = TransitionTable({})
        if transition_table:
            self._t_table = transition_table

    @property
    def t_table(self) -> TransitionTable:
        return self._t_table

    @property
    def current_state(self) -> State:
        return self._state

    def _undefined_decryptor(self, key: K, ctext: bytes) -> bytes:
        raise StateError("Method not allowed in this game")
        return (  # Compiler should know this is unreachable
            b"Does this ever return?"
            b" No, this never returns,"
            b" And its fate is still unlearned.",
        )

    @manage_state
    def initialize(self) -> None:
        """Initializes self by creating key and selecting b.

        Also clears an saved challenge ciphertexts.

        :raises StateError: if method called when disallowed.
        """

        """Challenger picks key and a b."""
        self._key = self._key_gen()
        self._b = secrets.choice([True, False])
        self._challenge_ctexts = set()

    @manage_state
    def encrypt_one(self, m0: bytes, m1: bytes) -> bytes:
        """Left-Right encryption oracle.

        Challenger encrypts m0 if b is False, else encrypts m1.

        :param m0: Left message
        :param m1: Right message
        :raise ValueError: if lengths of m0 and m1 are not equal.
        :raises StateError: if method called when disallowed.
        """

        # If these are None at this point, you've got a bad TransitionTable.
        cast(bool, self._b)
        cast(K, self._key)

        # Hmm, casts aren't working for me.
        assert self._key is not None

        if len(m0) != len(m1):
            raise ValueError("Message lengths must be equal")

        m = m1 if self._b else m0

        ctext = self._encryptor(self._key, m)
        if self.TRACK_CHALLENGE_CTEXTS:
            self._challenge_ctexts.add(hash_bytes(ctext))
        return ctext

    @manage_state
    def encrypt(self, ptext: bytes) -> bytes:
        """Encryption oracle.

        :param ptext: Message to be encrypted
        :raises StateError: if method called when disallowed.
        """

        assert self._key is not None
        return self._encryptor(self._key, ptext)

    @manage_state
    def decrypt(self, ctext: bytes) -> bytes:
        """Decryption oracle.

        :param ctext: Ciphertext to be decrypted
        :raises StateError: if method called when disallowed.
        """

        assert self._key is not None

        if hash_bytes(ctext) in self._challenge_ctexts:
            raise Exception(
                "Adversary is not allowed to call decrypt on challenge ctext"
            )

        return self._decryptor(self._key, ctext)

    @manage_state
    def finalize(self, guess: SupportsBool) -> bool:
        """
        True iff guess is the same as b of previously created challenger.

        :raises StateError: if method called when disallowed.
        """

        return guess == self._b


class IndCpa(Ind[K]):
    T_TABLE = TransitionTable(
        {
            State.STARTED: {Action.INITIALIZE: State.INITIALIZED},
            State.INITIALIZED: {Action.ENCRYPT_ONE: State.CHALLENGED},
            State.CHALLENGED: {
                Action.ENCRYPT_ONE: State.CHALLENGED,
                Action.FINALIZE: State.STARTED,
            },
        }
    )
    """Transition table for CPA game."""

    TRACK_CHALLENGE_CTEXTS = False
    """This game does not need to record challenge ctexts."""

    def __init__(
        self,
        key_gen: KeyGenerator[K],
        encryptor: Cryptor[K],
    ) -> None:
        """IND-CPA game.

        :param key_gen: A key generation function appropriate for encryptor
        :param encryptor:
            A function that takes a key and message and outputs ctext
        """

        super().__init__(key_gen=key_gen, encryptor=encryptor)
        self._t_table = self.T_TABLE


class IndEav(Ind[K]):
    T_TABLE = TransitionTable(
        {
            State.STARTED: {Action.INITIALIZE: State.INITIALIZED},
            State.INITIALIZED: {Action.ENCRYPT_ONE: State.CHALLENGED},
            State.CHALLENGED: {
                Action.FINALIZE: State.STARTED,
            },
        }
    )
    """Transition table for EAV game"""

    TRACK_CHALLENGE_CTEXTS = False
    """This game does not need to record challenge ctexts."""

    def __init__(
        self,
        key_gen: KeyGenerator[K],
        encryptor: Cryptor[K],
    ) -> None:
        """IND-EAV game.

        :param key_gen: A key generation function appropriate for encryptor
        :param encryptor:
            A function that takes a key and message and outputs ctext
        :raises StateError: if methods called in disallowed order.
        """

        super().__init__(key_gen=key_gen, encryptor=encryptor)
        self._t_table = self.T_TABLE


class IndCca2(Ind[K]):
    T_TABLE = TransitionTable(
        {
            State.STARTED: {Action.INITIALIZE: State.INITIALIZED},
            State.INITIALIZED: {
                Action.ENCRYPT_ONE: State.CHALLENGED,
                Action.ENCRYPT: State.INITIALIZED,
                Action.DECRYPT: State.INITIALIZED,
            },
            State.CHALLENGED: {
                Action.FINALIZE: State.STARTED,
                Action.ENCRYPT: State.CHALLENGED,
                Action.DECRYPT: State.CHALLENGED,
            },
        }
    )
    """Transition table for IND-CCA2 game"""

    TRACK_CHALLENGE_CTEXTS = True
    """CCA2 needs to prevent decrypt() from decrypting challenge ctexts."""

    def __init__(
        self,
        key_gen: KeyGenerator[K],
        encryptor: Cryptor[K],
        decrytpor: Cryptor[K],
    ) -> None:
        """IND-CCA2 game.

        :param key_gen: A key generation function appropriate for encryptor
        :param encryptor:
            A function that takes a key and message and outputs ctext
        :param decryptor:
            A function that takes a key and ciphertext and outputs plaintext
        :raises StateError: if methods called in disallowed order.
        """

        super().__init__(
            key_gen=key_gen, encryptor=encryptor, decryptor=decrytpor
        )
        self._t_table = self.T_TABLE


class IndCca1(Ind[K]):
    T_TABLE = TransitionTable(
        {
            State.STARTED: {Action.INITIALIZE: State.INITIALIZED},
            State.INITIALIZED: {
                Action.ENCRYPT_ONE: State.CHALLENGED,
                Action.ENCRYPT: State.INITIALIZED,
                Action.DECRYPT: State.INITIALIZED,
            },
            State.CHALLENGED: {
                Action.FINALIZE: State.STARTED,
                Action.ENCRYPT: State.CHALLENGED,
            },
        }
    )
    """Transition table for IND-CCA1 game"""

    TRACK_CHALLENGE_CTEXTS = False
    """This game does not need to record challenge ctexts."""

    def __init__(
        self,
        key_gen: KeyGenerator[K],
        encryptor: Cryptor[K],
        decrytpor: Cryptor[K],
    ) -> None:
        """IND-CCA1 game.

        :param key_gen: A key generation function appropriate for encryptor
        :param encryptor:
            A function that takes a key and message and outputs ctext
        :param decryptor:
            A function that takes a key and ciphertext and outputs plaintext
        :raises StateError: if methods called in disallowed order.
        """

        super().__init__(
            key_gen=key_gen, encryptor=encryptor, decryptor=decrytpor
        )
        self._t_table = self.T_TABLE



import math
import secrets
from bisect import bisect
from collections.abc import MutableSequence, Sequence
from itertools import accumulate, repeat
from typing import Any, Optional


def randrange(*args: int) -> int:
    """
    Like :py:func:`random.randrange`, but uses RNG from :py:mod:`secrets`.
    """

    if any([not isinstance(arg, int) for arg in args]):
        raise TypeError("arguments must be integers")

    start = 0
    step = 1
    match len(args):
        case 1:
            stop = args[0]
        case 2:
            start = args[0]
            stop = args[1]
        case 3:
            start = args[0]
            stop = args[1]
            step = args[2]

        case _:
            raise TypeError("Must have 1, 2, or 3 arguments")

    diff = stop - start
    if diff < 1:
        raise ValueError("stop must be greater than start")

    if step < 1:
        raise ValueError("step must be positive")

    if diff == 1:
        return start

    if step >= diff:  # only the bottom of the range will be allowed
        return start

    r = secrets.randbelow(diff // step)
    r *= step
    r += start

    return r


def shuffle(x: MutableSequence[Any]) -> None:
    """Like :py:func:`random.shuffle`, but uses RNG from :py:mod:`secrets`."""

    # Uses the "modern" Fisher-Yates shuffle from Knuth via
    # https://en.wikipedia.org/wiki/Fisher–Yates_shuffle#The_modern_algorithm

    n = len(x)
    if n < 2:
        return
    for i in range(n - 1):
        j = randrange(i, n)
        x[i], x[j] = x[j], x[i]


# from FullRandom example in
# https://docs.python.org/3/library/random.html#examples
def random() -> float:
    """returns a 32-bit float in [0.0, 1.0)"""

    mantissa = 0x10_0000_0000_0000 | secrets.randbits(52)
    exponent = -53
    x = 0
    while not x:
        x = secrets.randbits(32)
        exponent += x.bit_length() - 32
    return math.ldexp(mantissa, exponent)


def choices[T](
    population: Sequence[T],
    weights: Optional[Sequence[float]] = None,
    *,
    cum_weights: Optional[Sequence[float]] = None,
    k: int = 1,
) -> Sequence[T]:
    """Return a k sized list of population elements chosen with replacement.

    If the relative weights or cumulative weights are not specified,
    the selections are made with equal probability.

    This is the same as :py:func:`random.choices` except that it uses the
    the random number generator from :py:mod:1secrets`.
    Indeed, the implementation is just copied from that source.
    """

    # copied from https://github.com/python/cpython/blob/3.13/Lib/random.py#L458
    # except with static typing
    # and less abstract naming conventions for imported things
    n = len(population)
    if cum_weights is None:
        if weights is None:
            floor = math.floor
            return [population[floor(random() * n)] for i in repeat(None, k)]
        try:
            cum_weights = list(accumulate(weights))
        except TypeError:
            if not isinstance(weights, int):
                raise
            k = weights
            raise TypeError(
                f"The number of choices must be a keyword argument: {k=}"
            ) from None
    elif weights is not None:
        raise TypeError("Cannot specify both weights and cumulative weights")
    if len(cum_weights) != n:
        raise ValueError("The number of weights does not match the population")
    total = float(cum_weights[-1])
    if total <= 0.0:
        raise ValueError("Total of weights must be greater than zero")
    if not math.isfinite(total):
        raise ValueError("Total of weights must be finite")
    hi = n - 1
    return [
        population[bisect(cum_weights, random() * total, 0, hi)]
        for i in repeat(None, k)
    ]
